# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later
import logging
from datetime import datetime, timedelta
from gettext import gettext as _
from gettext import pgettext as C_
from gettext import ngettext as N_

from gi.repository import Adw, Gio, GObject, Gtk

from .. import shared  # type: ignore
from ..background_queue import (ActivityType, BackgroundActivity,
                                BackgroundQueue)
from ..models.movie_model import MovieModel
from ..models.series_model import SeriesModel
from ..providers.local_provider import LocalProvider as local
from ..providers.tmdb_provider import TMDBProvider as tmdb
from ..views.content_view import ContentView
from ..widgets.theme_switcher import ThemeSwitcher


@Gtk.Template(resource_path=shared.PREFIX + '/ui/views/main_view.ui')
class MainView(Adw.Bin):
    """
    This class represents the main view of the app.

    Properties:
        None

    Methods:
        refresh(): Causes the window to update its contents

    Signals:
        None
    """

    __gtype_name__ = 'MainView'

    _tab_stack = Gtk.Template.Child()
    _show_search_btn = Gtk.Template.Child()
    _menu_btn = Gtk.Template.Child()
    _banner = Gtk.Template.Child()
    _background_indicator = Gtk.Template.Child()
    _search_bar = Gtk.Template.Child()
    _search_mode = Gtk.Template.Child()
    _search_entry = Gtk.Template.Child()

    _needs_refresh = ''

    def __init__(self, window):
        super().__init__()
        self.app = window.app

        self._tab_stack.add_titled_with_icon(ContentView(movie_view=True),
                                             'movies',
                                             C_('Category', 'Movies'),
                                             'movies'
                                             )

        self._tab_stack.add_titled_with_icon(ContentView(movie_view=False),
                                             'series',
                                             C_('Category', 'TV Series'),
                                             'series'
                                             )

        shared.schema.bind('win-tab', self._tab_stack,
                           'visible-child-name', Gio.SettingsBindFlags.DEFAULT)
        shared.schema.bind('offline-mode', self._banner,
                           'revealed', Gio.SettingsBindFlags.GET)
        shared.schema.bind('search-enabled', self._search_bar,
                           'search-mode-enabled', Gio.SettingsBindFlags.DEFAULT)
        shared.schema.bind('search-enabled', self._show_search_btn,
                           'active', Gio.SettingsBindFlags.DEFAULT)

        self._search_mode.connect(
            'notify::selected', self._on_search_mode_changed)

        self._tab_stack.connect(
            'notify::visible-child-name', self._check_needs_refresh)

        # Theme switcher (Adapted from https://gitlab.gnome.org/tijder/blueprintgtk/)
        self._menu_btn.get_popover().add_child(ThemeSwitcher(), 'themeswitcher')

    def _check_needs_refresh(self, pspec: GObject.ParamSpec, user_data: object | None) -> None:
        """
        Checks if the tab switched to is pending a refresh and does it if needed.

        Args:
            pspec (GObject.ParamSpec): pspec of the changed property
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """
        if self._tab_stack.get_visible_child_name() == 'movies' and self._needs_refresh == 'movies':
            self._tab_stack.get_child_by_name('movies').refresh_view()
            logging.info('Refreshed movies tab')
            self._needs_refresh = ''
        elif self._tab_stack.get_visible_child_name() == 'series' and self._needs_refresh == 'series':
            self._tab_stack.get_child_by_name('series').refresh_view()
            logging.info('Refreshed TV Series tab')
            self._needs_refresh = ''

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for "map" signal.
        Calls method to check if an automatic content update is due.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        if not shared.schema.get_boolean('first-run'):
            self._check_update_content()

    def _check_update_content(self) -> None:
        """
        Checks if a content update is due, triggering it by adding background activities, if necessary.

        Args:
            None

        Returns:
            None
        """

        last_check = datetime.fromisoformat(
            shared.schema.get_string('last-update'))
        frequency = shared.schema.get_string('update-freq')
        last_notification_check = datetime.fromisoformat(
            shared.schema.get_string('last-notification-update'))

        if last_notification_check + timedelta(hours=12) < datetime.now():
            shared.schema.set_string(
                'last-notification-update', datetime.now().strftime('%Y-%m-%d %H:%M'))
            logging.info('Starting automatic notification list update...')
            BackgroundQueue.add(
                activity=BackgroundActivity(
                    activity_type=ActivityType.UPDATE,
                    title=C_('Notification List activity title',
                             'Automatic update of notification list'),
                    task_function=self._update_notification_list),
                on_done=self._on_notification_list_done)

        logging.debug(
            f'Last update done on {last_check}, frequency {frequency}')

        run = True
        match frequency:
            case 'day':
                if last_check + timedelta(days=1) < datetime.now():
                    logging.info('Starting automatic update...')
                    BackgroundQueue.add(
                        activity=BackgroundActivity(
                            activity_type=ActivityType.UPDATE,
                            title=C_('Background activity title',
                                     'Automatic update'),
                            task_function=self._update_content),
                        on_done=self._on_update_done)
            case 'week':
                if last_check + timedelta(days=7) < datetime.now():
                    logging.info('Starting automatic update...')
                    BackgroundQueue.add(
                        activity=BackgroundActivity(
                            activity_type=ActivityType.UPDATE,
                            title=C_('Background activity title',
                                     'Automatic update'),
                            task_function=self._update_content),
                        on_done=self._on_update_done)
            case 'month':
                if last_check + timedelta(days=30) < datetime.now():
                    logging.info('Starting automatic update...')
                    BackgroundQueue.add(
                        activity=BackgroundActivity(
                            activity_type=ActivityType.UPDATE,
                            title=C_('Background activity title',
                                     'Automatic update'),
                            task_function=self._update_content),
                        on_done=self._on_update_done)
            case 'never':
                return

        shared.schema.set_string(
            'last-update', datetime.now().strftime('%Y-%m-%d'))

    def _update_content(self, activity: BackgroundActivity) -> None:
        """
        Performs a content update on content added from TMDB.

        Args:
            activity (BackgroundActivity): the calling activity

        Returns:
            None
        """

        movies = local.get_all_movies()
        if movies:
            for movie in movies:    # type: ignore
                if not movie.manual:
                    new_movie = MovieModel(tmdb.get_movie(movie.id))
                    local.update_movie(old=movie, new=new_movie)

        series = local.get_all_series()
        if series:
            for serie in series:    # type: ignore
                if not serie.manual:
                    new_serie = SeriesModel(tmdb.get_serie(serie.id))
                    local.update_series(old=serie, new=new_serie)

    def _on_update_done(self,
                        source: GObject.Object,
                        result: Gio.AsyncResult,
                        cancellable: Gio.Cancellable,
                        activity: BackgroundActivity):
        """Callback to complete async activity"""

        self.refresh()
        logging.info('Automatic update done')
        activity.end()

    def _update_notification_list(self, activity: BackgroundActivity) -> None:
        """
        Performs a content update on the notification list.

        Args:
            activity (BackgroundActivity): the calling activity

        Returns:
            None
        """
        series = local.get_all_series_notification_list()

        new_release_series = []
        soon_release_series = []
        out_of_production_series = []

        for serie in series:

            last_air_date = datetime.strptime(serie.last_air_date, '%Y-%m-%d')

            # Get the latest info for the series from TMDB
            new_serie = SeriesModel(tmdb.get_serie(serie.id))
            new_last_air_date = datetime.strptime(
                new_serie.last_air_date, '%Y-%m-%d')
            if new_serie.next_air_date != '':
                new_next_air_date = datetime.strptime(
                    new_serie.next_air_date, '%Y-%m-%d')
            else:
                # create bogus next air date if it does not exist
                new_next_air_date = datetime.now() + timedelta(days=10)

            # Check if the latest release is newer than the last saved in the database -> new release has come out.
            if last_air_date < new_last_air_date:
                # Set the new release status and add the series to the new releases list and set soon_release to false
                local.set_new_release_status(serie.id, True)
                local.set_soon_release_status(serie.id, False)
                local.set_recent_change_status(serie.id, True)
                new_release_series.append(new_serie)
                # we only save one, since we do not use it if more than one series has a new release
                new_release_series_span = datetime.now() - new_last_air_date

            # Check if the next air date is set to soon (7 days in the future)
            if datetime.now() + timedelta(days=7) > new_next_air_date:
                local.set_soon_release_status(serie.id, True)
                # if we also detect a considerable amount of time bewteen epsidoe notify user that the series has new releases coming soon.
                # 3 weeks are chosen to include the new streaming release model of two chunks a month apart but not spam the user for weekly or bi-weekly releases
                if new_next_air_date - timedelta(days=20) > last_air_date:
                    local.set_recent_change_status(serie.id, True)
                    soon_release_series.append(new_serie)
                    soon_release_series_span = new_next_air_date - datetime.now()

            # Check if the series went from in production to not in production
            if serie.in_production == 1 and new_serie.in_production == 0:
                local.set_recent_change_status(serie.id, True)
                out_of_production_series.append(new_serie)
                local.set_notification_list_status(serie.id, False)

            # refetch serie to get all the correct flags that we set from the database
            serie = local.get_series_by_id(serie.id)
            local.update_series(serie, new_serie)

        movies = local.get_all_movies_notification_list()

        new_release_movies = []
        soon_release_movies = []

        for movie in movies:

            # Get the latest info for the movie from TMDB
            new_movie = MovieModel(tmdb.get_movie(movie.id))
            release_date = datetime.strptime(
                new_movie.release_date, '%Y-%m-%d')

            if release_date < datetime.now():
                local.set_recent_change_status(movie.id, True, movie=True)
                local.set_new_release_status(movie.id, True, movie=True)
                local.set_soon_release_status(movie.id, False, movie=True)
                if not movie.new_release:  # if new_release was not set send a notification
                    new_release_movies_span = datetime.now() - release_date
                    new_release_movies.append(new_movie)
            elif release_date < datetime.now() + timedelta(days=14):
                local.set_recent_change_status(movie.id, True, movie=True)
                local.set_soon_release_status(movie.id, True, movie=True)
                if not movie.soon_release:  # if soon_release was not set send a notification
                    soon_release_movies_span = release_date - datetime.now()
                    soon_release_movies.append(new_movie)
            # For movies we do not need to refetch the movie from the local db since the new data gets inserted by SQL UPDATE
            local.update_movie(movie, new_movie)

        def length_check(x): return len(x) > 0
        count = length_check(new_release_movies) + length_check(new_release_series) + length_check(
            soon_release_movies) + length_check(soon_release_series) + length_check(out_of_production_series)

        if count == 0:
            return
        elif count == 1:
            if new_release_series:
                if len(new_release_series) == 1:
                    # TRANSLATOR: {title} is the title of the series
                    title = _("New release for {title}").format(
                        title=new_release_series[0].title)
                    # TRANSLATOR: {title} is the title of the series and {days} the number of days
                    body = N_("A new episode of {title} was released {days} day ago.", "A new episode of {title} was released {days} days ago.",
                              new_release_series_span.days).format(title=new_release_series[0].title, days=new_release_series_span.days)
                else:
                    # TRANSLATOR: {num} is the number of TV Series
                    title = _("New release for {num} TV series on your watchlist").format(
                        num=len(new_release_series))
                    series = ", ".join(new.title for new in new_release_series)
                    # TRANSLATOR: {series} is a list of TV series seperated by a comma
                    body = _("The TV Series are {series}.").format(
                        series=series)

            if soon_release_series:
                if len(soon_release_series) == 1:
                    # TRANSLATOR: {title} is the title of the series with a new episode soon
                    title = _("{title} will have a release soon").format(
                        title=soon_release_series[0].title)
                    # TRANSLATOR: {title} is the title of the series and {days} the number of days
                    body = N_("A new episode will release in {days} day.", "A new episode will release in {days} days.", soon_release_series_span.days).format(
                        title=soon_release_series[0].title, days=soon_release_series_span.days)
                else:
                    # TRANSLATOR: {num} is the number of TV Series
                    title = _("{num} TV Series on your watchlist will have a new episode soon").format(
                        num=len(soon_release_series))
                    series = ", ".join(
                        soon.title for soon in soon_release_series)
                    # TRANSLATOR: {series} is a list of all series affected seperated by a comma
                    body = _("The TV Series are {series}.").format(
                        series=series)

            if out_of_production_series:
                if len(out_of_production_series) == 1:
                    # TRANSLATOR: {title} is the title of the series that has gone out of production
                    title = _("{title} has gone out of production").format(
                        title=out_of_production_series[0].title)
                    # TRANSLATOR: {title} is the title of the series that has gone out of production
                    body = _("{title} has wrapped up its run. Now is the perfect time to revisit your favorite moments or find the next binge!")
                else:
                    # TRANSLATOR: {num} is the number of TV Series
                    title = _("{num} TV Series of your watchlist have gone out of production").format(
                        num=len(out_of_production_series))
                    series = ", ".join(
                        out.title for out in out_of_production_series)
                    # TRANSLATOR: {series} is a list of all series affected seperated by a comma
                    body = _("The TV Series are {series}.").format(
                        series=series)

            if new_release_movies:
                if len(new_release_movies) == 1:
                    # TRANSLATOR: {title} is the title of the movie that has had its release
                    title = _("{title} has had its release!").format(
                        title=new_release_movies[0].title)
                    # TRANSLATOR: {title} is the title of the series and {days} the number of days
                    body = N_("{title} was released {days} day ago.", "{title} was released {days} days ago.", new_release_movies_span.days).format(
                        title=new_release_movies[0].title, days=new_release_movies_span.days)
                else:
                    title = _("{num} movies on your watchlist have had their releases.").format(
                        num=len(new_release_movies))
                    movies = ", ".join(new.title for new in new_release_movies)
                    # TRANSLATOR: {movies} is a list of all movies affected seperated by a comma
                    body = _("The movies are {movies}.").format(movies=movies)

            if soon_release_movies:
                if len(soon_release_movies) == 1:
                    # TRANSLATOR: {title} is the title of the movie that will have its release soon
                    title = _("{title} will have its release soon!").format(
                        title=soon_release_movies[0].title)
                    # TRANSLATOR: {title} is the title of the movie and {days} the number of days
                    body = N_("{title} will have its release in {days} day.", "{title} will have its release in {days} days.", soon_release_movies_span.days).format(
                        title=soon_release_movies[0].title, days=soon_release_movies_span.days)
                else:
                    # TRANSLATOR: {num} is the number of movies
                    title = _("{num} movies on your watchlist will have their releases soon")
                    movies = ", ".join(soon.title for soon in soon_release_movies)
                    # TRANSLATOR: {movies} will be list of all series affected seperated by a comma
                    body = _("The movies are {movies}.").format(movies=movies)

            notification = Gio.Notification.new(title)
            notification.set_body(body)
            self.app.send_notification(None, notification)
        else:
            count_movies = len(new_release_movies) + len(soon_release_movies)
            count_series = len(
                new_release_series) + len(soon_release_series) + len(out_of_production_series)
            # TRANSLATOR: {num} is the number of affected items
            title = _("{num} items of your watchlist have an update").format(
                num=count_movies + count_series)

            if count_movies > 0 and count_series > 0:
                # TRANSLATOR: {count_movies} is the number of movies
                movie = N_("These updates affect {count_movies} movie", "These updates affect {count_movies} movies", count_movies)
                # TRANSLATOR: the connector between the two parts of the sentence (foo and bar)
                connector = _(" and ")
                # TRANSLATOR: {count_series} is the number of series
                series = N_(" {count_series} TV serie", " {count_series} TV series", count_series)
                
                body = f"{movie}{connector}{series}."
            elif count_movies > 0:
                # TRANSLATOR: {count_movies} is the number of movies
                body = N_("These updates affect {count_movies} movie", "These updates affect {count_movies} movies", count_movies)
            elif count_series > 0:
                # TRANSLATOR: {count_series} is the number of series
                body = N_("These updates affect {count_series} TV serie", "These updates affect {count_series} TV series", count_series)

            notification = Gio.Notification.new(title)
            notification.set_body(body)
            self.app.send_notification(None, notification)

    def _on_notification_list_done(self,
                                   source: GObject.Object,
                                   result: Gio.AsyncResult,
                                   cancellable: Gio.Cancellable,
                                   activity: BackgroundActivity):
        """Callback to complete async activity"""

        self.refresh()
        logging.info('Automatic notification list update done')
        activity.end()

    def refresh(self) -> None:
        """
        Refreshes the visible window.

        Args:
            None

        Returns:
            None
        """

        if self._tab_stack.get_visible_child_name() == 'movies':
            self._tab_stack.get_child_by_name('movies').refresh_view()
            logging.info('Refreshed movies tab')
            self._needs_refresh = 'series'
        else:
            self._tab_stack.get_child_by_name('series').refresh_view()
            logging.info('Refreshed TV series tab')
            self._needs_refresh = 'movies'

    @Gtk.Template.Callback()
    def _on_searchentry_search_changed(self, user_data: GObject.GPointer | None) -> None:
        shared.schema.set_string('search-query', self._search_entry.get_text())

    @Gtk.Template.Callback()
    def _on_search_btn_toggled(self, user_data: GObject.GPointer | None) -> None:
        shared.schema.set_boolean(
            'search-enabled', self._show_search_btn.get_active())
        shared.schema.set_string('search-query', '')

    def _on_search_mode_changed(self, pspec: GObject.ParamSpec, user_data: object | None) -> None:
        if self._search_mode.get_selected() == 0:
            shared.schema.set_string('search-mode', 'title')
        else:
            shared.schema.set_string('search-mode', 'genre')



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

from enum import Enum
from typing import Callable

from gi.repository import Gio, GLib, GObject

from . import shared  # type: ignore


class ActivityType(Enum):
    """
    Enum for the types of background activities
    """

    ADD = 0
    REMOVE = 1
    UPDATE = 2


class BackgroundActivity(GObject.GObject):
    """
    An activity that is run in the background.

    Properties:
        title (str): a title
        activity_type (str): an activity type, name as in ActivityType
        callback (callable): a function to run in the background
        completed (bool): indicates if the activity is completed

    Methods:
        start(): runs self.callback in a separate thread
        end(): marks the activity as completed

    Signals:
        None
    """

    __gtype_name__ = 'BackgroundActivity'

    title = GObject.Property(type=str, default='')
    activity_type = GObject.Property(type=str, default='')
    task_function = GObject.Property(type=object, default=None)
    completed = GObject.Property(type=bool, default=False)
    has_error = GObject.Property(type=bool, default=False)

    def __init__(self, activity_type: ActivityType, title: str = '', task_function: Callable | None = None):
        super().__init__()
        self.activity_type = activity_type.name
        self.title = title
        self.task_function = task_function
        self._cancellable = Gio.Cancellable()

    def start(self, on_done: Callable) -> None:
        """
        Runs self.callback in a separate thread. The callback must call end() to mark the activity as completed.

        Args:
            None

        Returns:
            None
        """

        task = Gio.Task.new(self, None, on_done, self._cancellable, self)
        task.set_return_on_cancel(True)
        task.run_in_thread(self._run_in_thread)

    def _run_in_thread(self,
                       task: Gio.Task,
                       source_object: GObject.Object,
                       task_data: object | None,
                       cancelable: Gio.Cancellable):
        """Callback to run self.task_function in a thread"""

        if task.return_error_if_cancelled():
            return
        outcome = self.task_function(self)  # type: ignore
        task.return_value(outcome)

    def activity_finish(self, result: Gio.AsyncResult, caller: GObject.Object):
        """
        Completes the async operation and marks the activity as completed.

        Args:
            None

        Returns:
            None
        """

        if not Gio.Task.is_valid(result, caller):
            return -1

        # self.completed = True
        return result.propagate_value().value

    def end(self) -> None:
        """
        Marks the activity as completed.

        Args:
            None

        Returns:
            None
        """

        self.completed = True

    def error(self) -> None:
        """
        Marks the activity with an error.

        Args:
            None

        Returns:
            None
        """

        self.has_error = True


class BackgroundQueue:
    """
    A queue of background activities.

    Properties:
        None

    Methods:
        add(activity: BackgroundActivity): adds an activity to the queue
        get_queue(): returns the queue

    Signals:
        None
    """

    _queue = Gio.ListStore.new(item_type=BackgroundActivity)

    @staticmethod
    def add(activity: BackgroundActivity, on_done: Callable) -> None:
        """
        Adds an activity to the queue and starts its execution.

        Args:
            activity (BackgroundActivity): the activity to add

        Returns:
            None
        """

        BackgroundQueue._queue.append(activity)
        activity.start(on_done)

    @staticmethod
    def get_queue() -> Gio.ListStore:
        """
        Returns the queue

        Args:
            None

        Returns:
            the queue (a Gio.ListStore)
        """

        return BackgroundQueue._queue



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import glob
import os
import re

import requests
from gi.repository import GObject
from PIL import Image

from .. import shared  # type: ignore


class EpisodeModel(GObject.GObject):
    """
    This class represents an episode object stored in the db.

    Properties:
        id (str): episode id
        number (int): episode number in its season
        overview (str): episode overview
        runtime (int): episode runtime in minutes
        season_number (int): season the episode belongs to
        show_id (int): id of the show the episode belongs to
        still_path (str): uri of the episode still
        title (str): episode title
        watched (bool): whether the episode has been watched or not

    Methods:
        None

    Signals:
        None
    """

    __gtype_name__ = 'EpisodeModel'

    id = GObject.Property(type=str, default='')
    number = GObject.Property(type=int, default=0)
    overview = GObject.Property(type=str, default='')
    runtime = GObject.Property(type=int, default=0)
    season_number = GObject.Property(type=int, default=0)
    show_id = GObject.Property(type=str, default='')
    still_path = GObject.Property(type=str, default='')
    title = GObject.Property(type=str, default='')
    watched = GObject.Property(type=bool, default=False)

    def __init__(self, d=None, t=None):
        super().__init__()

        if d is not None:
            self.id = d['id']
            self.number = d['episode_number']
            self.overview = re.sub(r'\s{2}', ' ', d['overview'])
            self.runtime = d['runtime'] if d['runtime'] else 0
            self.season_number = d['season_number']
            self.show_id = d['show_id']
            self.still_path = self._download_still(d['still_path'])
            self.title = d['name']
            self.watched = False
        else:
            self.id = t[0]  # type: ignore
            self.number = t[1]  # type: ignore
            self.overview = t[2]  # type: ignore
            self.runtime = t[3]  # type: ignore
            self.season_number = t[4]  # type: ignore
            self.show_id = t[5]  # type: ignore
            self.still_path = t[6]  # type: ignore
            self.title = t[7]  # type: ignore
            self.watched = t[8]  # type: ignore

    def _download_still(self, path: str) -> str:
        """
        Returns the uri of the still image on the local filesystem, downloading if necessary.

        Args:
            path (str): path to dowload from

        Returns:
            str with the uri of the still image
        """

        if not path:
            return f'resource://{shared.PREFIX}/blank_still.jpg'

        if not os.path.exists(f'{shared.series_dir}/{self.show_id}/{self.season_number}'):
            os.makedirs(
                f'{shared.series_dir}/{self.show_id}/{self.season_number}')

        files = glob.glob(
            f'{path[1:-4]}.jpg', root_dir=f'{shared.series_dir}/{self.show_id}/{self.season_number}')
        if files:
            return f'file://{shared.series_dir}/{self.show_id}/{self.season_number}/{files[0]}'

        url = f'https://image.tmdb.org/t/p/w500{path}'
        try:
            r = requests.get(url)
            if r.status_code == 200:
                with open(f'{shared.series_dir}/{self.show_id}/{self.season_number}{path}', 'wb') as f:
                    f.write(r.content)

                with Image.open(f'{shared.series_dir}/{self.show_id}/{self.season_number}{path}') as img:
                    img = img.resize((500, 281))
                    img.save(
                        f'{shared.series_dir}/{self.show_id}/{self.season_number}{path}', 'JPEG')
                return f'file://{shared.series_dir}/{self.show_id}/{self.season_number}{path}'
            else:
                return f'resource://{shared.PREFIX}/blank_still.jpg'
        except (requests.exceptions.ConnectionError, requests.exceptions.SSLError):
            return f'resource://{shared.PREFIX}/blank_still.jpg'



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os
import shutil
from datetime import date, datetime
from gettext import gettext as _
from gettext import pgettext as C_
from typing import List, Tuple
from urllib.parse import unquote
from PIL import Image, ImageStat
from pathlib import Path

from gi.repository import Adw, Gio, GLib, GObject, Gtk

from .. import shared  # type: ignore
from ..background_queue import (ActivityType, BackgroundActivity,
                                BackgroundQueue)
from ..dialogs.edit_season_dialog import EditSeasonDialog
from ..models.episode_model import EpisodeModel
from ..models.language_model import LanguageModel
from ..models.movie_model import MovieModel
from ..models.season_model import SeasonModel
from ..models.series_model import SeriesModel
from ..providers.local_provider import LocalProvider as local
from ..widgets.season_expander import SeasonExpander


@Gtk.Template(resource_path=shared.PREFIX + '/ui/dialogs/add_manual.ui')
class AddManualDialog(Adw.Dialog):
    """
    This class represents the dialog to manually add content to the db.

    Properties:
        edit_mode (bool): whether or not the window is in add/edit mode

    Methods:
        update_seasons_ui(): rebuilds the ui to reflect changes to seasons
        get_season(title: str, poster: str, episodes: list): returns the tuple with the provided data or an empty tuple

    Signals:
        edit-saved(SeriesModel or MovieModel): emited when the user clicks the save button
    """

    __gtype_name__ = 'AddManualDialog'

    edit_mode = GObject.Property(type=bool, default=False)

    _movies_btn = Gtk.Template.Child()
    _series_btn = Gtk.Template.Child()
    _save_btn = Gtk.Template.Child()
    _poster = Gtk.Template.Child()
    _title_entry = Gtk.Template.Child()
    _release_date_menu_btn = Gtk.Template.Child()
    _calendar = Gtk.Template.Child()
    _genres_entry = Gtk.Template.Child()
    _runtime_spinrow = Gtk.Template.Child()
    _tagline_entry = Gtk.Template.Child()
    _creator_entry = Gtk.Template.Child()
    _overview_text = Gtk.Template.Child()
    _seasons_group = Gtk.Template.Child()
    _status_entry = Gtk.Template.Child()
    _original_language_comborow = Gtk.Template.Child()
    _language_model = Gtk.Template.Child()
    _original_title_entry = Gtk.Template.Child()
    _budget_spinrow = Gtk.Template.Child()
    _revenue_spinrow = Gtk.Template.Child()
    _production_checkbtn = Gtk.Template.Child()

    __gsignals__ = {
        'edit-saved': (GObject.SIGNAL_RUN_FIRST, None, (object,)),
    }

    seasons: list = []

    def __init__(self,
                 edit_mode: bool = False,
                 content: MovieModel | SeasonModel | None = None
                 ):
        super().__init__()

        self.edit_mode = edit_mode
        self._content = content

        logging.info(f'Edit mode: {self.edit_mode}')
        if edit_mode and type(self._content) is MovieModel:
            self.set_title(_('Edit Movie'))
            logging.debug(f'Editing [Movie] {self._content.title}')
        elif edit_mode and type(self._content) is SeriesModel:
            self.set_title(_('Edit TV Series'))
            logging.debug(f'Editing [TV Series] {self._content.title}')

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for the "map" signal.
        Setup the window by selecting the initial view (movies or tv series), the image for the poster and populates the language dropdown and the selected date for the release.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        self._overview_text.remove_css_class('view')

        languages = local.get_all_languages()
        # move 'no language' to 1st place
        languages.insert(0, languages.pop(len(languages)-6))
        for language in languages:
            self._language_model.append(language.name)

        self._poster.set_blank_image(
            f'resource://{shared.PREFIX}/blank_poster.jpg')

        if not self._content:
            if shared.schema.get_string('win-tab') == 'movies':
                self._movies_btn.set_active(True)
            else:
                self._series_btn.set_active(True)

            self._release_date_menu_btn.set_label(
                self._calendar.get_date().format('%x'))
            return

        # Both Movies and TV Series
        self._poster.set_image(self._content.poster_path)
        self._title_entry.set_text(self._content.title)
        self._title_entry.grab_focus()
        self._release_date_menu_btn.set_label(
            date.fromisoformat(self._content.release_date).strftime('%x'))
        self._calendar.select_day(GLib.DateTime.new_from_iso8601(
            datetime.fromisoformat(self._content.release_date).isoformat()+'Z'))
        self._genres_entry.set_text(', '.join(self._content.genres))
        self._tagline_entry.set_text(self._content.tagline)
        self._overview_text.get_buffer().set_text(self._content.overview, -1)
        self._status_entry.set_text(self._content.status)
        self._original_language_comborow.set_selected(
            self._get_selected_language_index(self._content.original_language))
        self._original_title_entry.set_text(self._content.original_title)

        # Movies specific
        if type(self._content) is MovieModel:
            self._movies_btn.set_active(True)
            self._runtime_spinrow.set_value(self._content.runtime)
            self._budget_spinrow.set_value(self._content.budget)
            self._revenue_spinrow.set_value(self._content.revenue)

        # TV Series specific
        if type(self._content) is SeriesModel:
            self._series_btn.set_active(True)
            self._creator_entry.set_text(','.join(self._content.created_by))
            self._production_checkbtn.set_active(self._content.in_production)
            self.seasons = self._parse_seasons(self._content.seasons)
            self.update_seasons_ui()

    def _parse_seasons(self, seasons_as_model: List[SeasonModel]) -> List[tuple]:
        """
        Parse seasons passed as a list of SeasonModels into a list of tuples.

        Args:
            seasons_as_model(List[SeasonModel]): a list of SeasonModels

        Returns:
            a list of tuples representing the same data
        """

        seasons_as_tuple = []
        for season in seasons_as_model:

            episodes_as_tuple = []
            for episode in season.episodes:
                episodes_as_tuple.append((
                    episode.title,
                    episode.number,
                    episode.runtime,
                    episode.overview,
                    episode.still_path,
                    episode.watched,
                    episode.id
                ))

            seasons_as_tuple.append((
                season.title,
                season.poster_path,
                episodes_as_tuple
            ))
        return seasons_as_tuple

    def _get_selected_language_index(self, language: LanguageModel) -> int:
        """
        Loops languages looking for the index of the provided language.

        Args:
            language (LanguageModel): language to look for

        Returns:
            index (int)
        """

        for idx in range(0, self._language_model.get_n_items()):
            if language.name == self._language_model.get_string(idx):
                return idx
        return -1

    @Gtk.Template.Callback('_on_title_changed')
    def _on_title_changed(self, user_data: object | None) -> None:
        """
        Callback for "changed" signal.
        Wrapper around self._enable_save_btn().

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        self._enable_save_btn()

    @Gtk.Template.Callback('_on_movies_btn_toggled')
    def _on_movies_btn_toggled(self, user_data: object | None) -> None:
        """
        Callback for "toggled" signal.
        Wrapper around self._enable_save_btn().

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        self._enable_save_btn()

    def _enable_save_btn(self) -> None:
        """
        Checks whether the "save" button should be active or not.

        Args:
            None

        Returns:
            None
        """

        # Movies: title required
        if self._movies_btn.get_active() and self._title_entry.get_text():
            self._save_btn.set_sensitive(True)
            return

        # TV Series: title and at least a season required
        if self._series_btn.get_active() and self._title_entry.get_text() and len(self.seasons) > 0:
            self._save_btn.set_sensitive(True)
            return

        self._save_btn.set_sensitive(False)

    @Gtk.Template.Callback('_on_calendar_day_selected')
    def _on_calendar_day_selected(self, user_data: object | None) -> None:
        """
        Callback for the "day-selected" signal.
        Sets the button's label as the selected date (user locale format)

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        self._release_date_menu_btn.set_label(
            self._calendar.get_date().format('%x'))

    @Gtk.Template.Callback('_on_season_add_btn_clicked')
    def _on_season_add_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for the "clicked" signal.
        Opens the "edit season" dialog.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        # TRANSLATORS: {num} is the season number
        logging.debug('Edit season dialog open')
        dialog = EditSeasonDialog(self, title=_(
            'Season {num}').format(num=len(self.seasons)+1))
        dialog.connect('edit-saved', self._on_edit_saved)
        dialog.present(dialog.parent)

    def _on_edit_saved(self, source: Gtk.Widget, title: str, poster_uri: str, episodes: List[tuple]) -> None:
        """
        Callback for the "edit-saved" signal.
        Appends the recieved data as a tuple in the seasons list and updates the ui.

        Args:
            source (Gtk.Widget): caller widget
            title (str): season title
            poster_uri (str): season poster uri
            episodes (List[tuple]): season episodes

        Returns:
            None
        """

        logging.info(f'Updated {title}, {poster_uri}, {episodes}')
        self.seasons.append((title, poster_uri, episodes))
        self.update_seasons_ui()

    @Gtk.Template.Callback('_on_done_btn_clicked')
    def _on_done_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for the "clicked" signal.
        Adds a background activity to add/update the content and closes the window.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        if not self.edit_mode:
            BackgroundQueue.add(
                activity=BackgroundActivity(  # TRANSLATORS: {title} is the content's title
                    activity_type=ActivityType.ADD,
                    title=C_('Background activity title', 'Add {title}').format(title=self._title_entry.get_text()), task_function=self._add_content_to_db),
                on_done=self._on_add_done)
        else:
            BackgroundQueue.add(
                activity=BackgroundActivity(  # TRANSLATORS: {title} is the content's title
                    activity_type=ActivityType.UPDATE,
                    title=C_('Background activity title', 'Update {title}').format(title=self._title_entry.get_text()), task_function=self._add_content_to_db),
                on_done=self._on_add_done)

        self.close()

    def _add_content_to_db(self, activity: BackgroundActivity) -> None:
        """
        Copies the poster image to the data folder, saves the content, and refreshes the main window.

        Args:
            activity (BackgroundActivity): the calling activity

        Returns:
            None
        """
        
       
        poster_uri, color = self._copy_image_to_data(self._poster.get_uri(),
                                              shared.poster_dir,
                                              self._title_entry.get_text()
                                              ) if not self.edit_mode else (self._content.poster_path, False)  # type: ignore
        if self._movies_btn.get_active():
            self._save_movie(poster_uri, color)
        else:
            self._save_series(poster_uri, color)

    def _on_add_done(self,
                     source: GObject.Object,
                     result: Gio.AsyncResult,
                     cancellable: Gio.Cancellable,
                     activity: BackgroundActivity):
        """Callback to complete async activity"""

        self.get_ancestor(Adw.ApplicationWindow).activate_action('win.refresh', None)
        activity.end()

    def _save_movie(self, poster_uri: str, color: bool) -> None:
        """
        Creates a MovieModel with the provided data and saves or updates the movie in the local db.

        Args:
            poster_uri (str): movie poster uri

        Returns:
            None
        """

        buffer = self._overview_text.get_buffer()
        start_iter = buffer.get_start_iter()
        end_iter = buffer.get_end_iter()
        overview = buffer.get_text(start_iter, end_iter, False)

        movie = MovieModel(t={
            "add_date": datetime.now(),  # add date
            "backdrop_path": '',  # background
            "budget":int(self._budget_spinrow.get_value()),  # budget
            "color": color,
            "genres": ''.join(self._genres_entry.get_text().split()),  # genres
            "id": local.get_next_manual_movie() if not self.edit_mode else self._content.id,  # id
            "manual":True,   # manual
            "original_language":local.get_language_by_name(self._original_language_comborow.get_selected_item(
            ).get_string()).iso_name,  # original language
            "original_title": self._original_title_entry.get_text(),  # original title
            "overview":overview,   # overview
            "poster_path": poster_uri,  # poster
            "release_date": self._calendar.get_date().format('%Y-%m-%d'),   # release date
            "revenue":int(self._revenue_spinrow.get_value()),  # revenue
            "runtime":int(self._runtime_spinrow.get_value()),  # runtime
            "status":self._status_entry.get_text(),  # status
            "tagline":self._tagline_entry.get_text(),  # tagline
            "title":self._title_entry.get_text(),   # title
            "watched": False if not self.edit_mode else self._content.watched  # watched
        })

        if not self.edit_mode:
            local.add_movie(movie=movie)
        else:
            local.update_movie(old=self._content, new=movie)
            self.emit('edit-saved', movie)

    def _save_series(self, series_poster_uri: str, color: bool) -> None:
        """
        Creates a SeriesModel with associated SeasonModels/EpisodeModels with the provided data and saves or updates the TV series in the local db.

        Args:
            poster_uri (str): tv series poster uri

        Returns:
            None
        """

        show_id = local.get_next_manual_series()
        base_season_id = local.get_next_manual_season()
        base_episode_id = local.get_next_manual_episode()

        seasons = []
        for idx, season in enumerate(self.seasons):

            # Create folder to store the images, if needed
            if not os.path.exists(f"{shared.series_dir}/{show_id}/{self._increment_manual_id(base_season_id, idx)}"):
                os.makedirs(
                    f"{shared.series_dir}/{show_id}/{self._increment_manual_id(base_season_id, idx)}")

            # Copy the season poster
            poster_uri, color = self._copy_image_to_data(season[1],
                                                  f"{shared.series_dir}/{show_id}/{self._increment_manual_id(base_season_id, idx)}",
                                                  season[0])

            episodes = []
            for jdx, episode in enumerate(season[2]):

                # Copy the episode still
                still_uri, color = self._copy_image_to_data(episode[4],
                                                     f"{shared.series_dir}/{show_id}/{self._increment_manual_id(base_season_id, idx)}",
                                                     episode[0]
                                                     )
                episode_id = self._increment_manual_id(base_episode_id, jdx)
                season_number = idx+1
                watched = False

                episodes.append(EpisodeModel(t=(
                    episode_id,     # id
                    episode[1],     # episode number
                    episode[3],     # overview
                    episode[2],     # runtime
                    season_number,  # season number
                    show_id,        # show id
                    still_uri,      # still uri
                    episode[0],     # title
                    watched         # watched
                )))

            base_episode_id = self._increment_manual_id(
                base_episode_id, len(episodes)+1)
            season_id = self._increment_manual_id(base_season_id, idx)
            season_number = idx+1

            seasons.append(SeasonModel(t=(
                len(episodes),  # episodes number
                season_id,      # id
                season_number,  # season number
                '',             # overview
                poster_uri,     # season poster
                season[0],      # title
                show_id,        # show id
                episodes        # season episodes
            )))

        buffer = self._overview_text.get_buffer()
        start_iter = buffer.get_start_iter()
        end_iter = buffer.get_end_iter()
        overview = buffer.get_text(start_iter, end_iter, False)

        serie = SeriesModel(t={
            "activate_notification": False,                                   # activate notification        
            "add_date": datetime.now(),                                       # add date
            "backdrop_path":'',                                               # backgroud
            "color": color,                                       # color
            "created_by": self._creator_entry.get_text(),                     # created by
            "episodes_number": self._compute_episode_number(seasons),         # episode number
            "genres": ''.join(self._genres_entry.get_text().split()),         # genres
            "id": show_id,                                                    # id
            "in_production": self._production_checkbtn.get_active(),          # in production
            "last_air_date": '',                                              # last air date
            "manual": True,                                                   # manual
            "new_release": False,                                             # new release flag
            "next_air_date": '',                                              # next air date           
            "original_language": local.get_language_by_name(self._original_language_comborow.get_selected_item(
            ).get_string()).iso_name,         # type: ignore                  # original language
            "original_title": self._original_title_entry.get_text(),          # original title
            "overview": overview,                                             # overview
            "poster_path": series_poster_uri,                                 # poster
            "release_date": self._calendar.get_date().format('%Y-%m-%d'),     # release date
            "seasons_number": len(seasons),                                   # seasons number
            "soon_release" :False,                                            # soon_release flag      
            "status": self._status_entry.get_text(),                          # status
            "tagline": self._tagline_entry.get_text(),                        # tagline
            "title": self._title_entry.get_text(),                            # title
            "watched": False,                                                 # watched
            "last_episode_number": 0,                                         # last episode number
            "recent_change": False,                                           # recent change flag
            "notes": '',                                                      # notes
            "seasons": seasons                                                # seasons
    })

        if self.edit_mode:
            local.delete_series(self._content.id)   # type: ignore
            local.add_series(serie=serie)
            self.emit('edit-saved', serie)
        else:
            local.add_series(serie=serie)

    def _increment_manual_id(self, id: str, amount: int = 1) -> str:
        """
        Increments the integer part of a manual id by amount.

        Args:
            id (str): id to increment
            amount (int): how much to increment by

        Returns:
            the incremented id
        """

        tmp = id.split('-')
        return f'M-{int(tmp[1]) + amount}'

    def _compute_badge_color(self, poster_path) -> bool:

        im = Image.open(poster_path)
        box = (im.size[0]-175, 0, im.size[0], 175)
        region = im.crop(box)
        median = ImageStat.Stat(region).median
        if sum(median) < 3 * 128:
            return True
        else:
            return False

    def _compute_episode_number(self, seasons: List[SeasonModel]) -> int:
        """
        Counts the total number of episodes of a tv series.

        Args:
            seasons (List[SeasonModel]): seasons to loop on

        Returns:
            the total number of episodes
        """

        num = 0
        for season in seasons:
            for episode in season.episodes:
                num += 1
        return num

    def _copy_image_to_data(self, src_uri: str, dest_folder: str, filename: str) -> Tuple[str,bool]:
        """
        Copies src_uri to dest_folder as filename. If src_uri is a resource (empty poster/still) the operation is not
        carried out.
        dest_folder must already exist on file system.

        Args:
            src_uri (str): source file uri
            dest_folder (str): path to the destination folder
            filename (str): new name for the copied file

        Returns:
            uri of the copied file or src_uri if is a resource.
        """
        
        if src_uri.startswith('file'):
            extension = src_uri[src_uri.rindex('.'):]
            dest_path = f"{dest_folder}/{unquote(filename)}{extension}"
            shutil.copy2(src_uri[7:], dest_path)
            return f"file://{dest_path}", self._compute_badge_color(Path(dest_path))
        return src_uri, False

    def update_seasons_ui(self) -> None:
        """
        Rebuilds the ui to reflect changes to seasons.

        Args:
            None

        Returns:
            None
        """

        # Empty PreferencesGroup
        list_box = self._seasons_group.get_first_child(
        ).get_last_child().get_first_child()  # ugly workaround
        list_box.remove_all()

        # Fill PreferencesGroup
        for season in self.seasons:
            self._seasons_group.add(SeasonExpander(
                season_title=season[0], poster_uri=season[1], episodes=season[2]))

        self._enable_save_btn()

    def get_season(self,
                   title: str,
                   uri: str,
                   episodes: List[tuple]) -> tuple:
        """
        Compares and retrieves the tuple containing the passed data.

        Args:
            title (str): a title
            uri (str): an uri
            episodes (List[tuple]): a list of tuples representing episodes

        Returns:
            tuple matching the passed data
        """

        for season in self.seasons:
            if (season[0] == title and
                season[1] == uri and
                    season[2] == episodes):
                return season
        return ()



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject


class LanguageModel(GObject.GObject):
    """
    This class represents a language object stored in the db.

    Properties:
        iso_name (str): ISO_639_1 code
        name (str): localized or English name

    Methods:
        None

    Signals:
        None
    """

    __gtype_name__ = 'LanguageModel'

    iso_name = GObject.Property(type=str, default='')
    name = GObject.Property(type=str, default='')

    def __init__(self, d=None, t=None):
        super().__init__()

        if d is not None:
            self.iso_name = d['iso_639_1']

            if d['name']:
                self.name = d['name']
            else:
                self.name = d['english_name']
        else:
            self.iso_name = t[0]  # type: ignore
            self.name = t[1]  # type: ignore



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from gi.repository import Adw, Gio, GLib, GObject, Gtk

from .. import shared  # type: ignore
from ..models.search_result_model import SearchResultModel
from ..providers.tmdb_provider import TMDBProvider


@Gtk.Template(resource_path=shared.PREFIX + '/ui/dialogs/add_tmdb.ui')
class AddTMDBDialog(Adw.Dialog):
    """
    This class represents the dialog used to search for movies and tv-series on TMDB.

    Properties:
        None

    Methods:
        None

    Signals:
        None
    """

    __gtype_name__ = 'AddTMDBDialog'

    _search_entry = Gtk.Template.Child()
    _stack = Gtk.Template.Child()
    _model = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

    @Gtk.Template.Callback('_on_searchentry_search_changed')
    def _on_searchentry_search_changed(self, user_data: object | None) -> None:
        """
        Callback for the "seach-changed" signal.
        Updates the GtkListModel used by the factory to populate the GtkListView.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        logging.info(f'Search query: "{self._search_entry.get_text()}"')

        if self._model.get_property('n-items') > 0:
            self._model.remove_all()

        if not self._search_entry.get_text():
            self._stack.set_visible_child_name('empty')
            return

        response = TMDBProvider().search(query=self._search_entry.get_text())
        if not response['results']:
            self._stack.set_visible_child_name('no-results')
            logging.info('No results for query')
            return

        for result in response['results']:
            if result['media_type'] in ['movie', 'tv']:
                search_result = SearchResultModel(result)
                logging.info(
                    f'Found [{"movie" if search_result.media_type == "movie" else "TV series"}] {search_result.title}, {search_result.year}')
                self._model.append(search_result)

        self._stack.set_visible_child_name('results')



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import glob
import re
from datetime import datetime, timedelta
from typing import List, Tuple
from pathlib import Path

import requests
from gi.repository import GLib, GObject
from PIL import Image, ImageFilter, ImageStat

import src.providers.local_provider as local

from .. import shared  # type: ignore
from ..models.language_model import LanguageModel
from ..models.season_model import SeasonModel


class SeriesModel(GObject.GObject):
    """
    This class represents a series object stored in the db.

    Properties:
        add_date (str): date of addition to the db (ISO format)
        backdrop_path (str): uri of the background image
        color (bool): color of the poster badges, False being dark
        created_by (List[str]): list of creators
        episodes_number (int): number of total episodes
        genres (List[str]): list of genres
        id (str): series id
        in_production (bool): whether the series is still in production
        last_air_date (str): the date of the last aired episode
        manual (bool): if the series is added manually
        new_release (bool): whether the has seen a new release since the last check
        next_air_date (str) : used to notify user on upcoming releases
        original_language (LanguageModel): LanguageModel of the original language
        original_title (str): series title in original language
        overview (str): series overview
        poster_path (str): uri of the poster image
        recent_change (bool): indicates that new/soon_release has chagned in the last check
        release_date (str): first air date in YYYY-MM-DD format
        seasons_number (int): number of total seasons
        seasons (List[SeasonModel]): list of SeasonModels
        status (str): series status
        tagline (str): series tagline
        title (str): series title
        watched (bool): whether the series has been watched completely or not
        activate_notification (bool): whether the series should be checked for new releases
        notes (str): user notes

    Methods:
        None

    Signals:
        None
    """

    __gtype_name__ = 'SeriesModel'
    activate_notification = GObject.Property(type=bool, default=False)
    add_date = GObject.Property(type=str, default='')
    backdrop_path = GObject.Property(type=str, default='')
    color = GObject.Property(type=bool, default=True)
    created_by = GObject.Property(type=GLib.strv_get_type())
    episodes_number = GObject.Property(type=int, default=0)
    genres = GObject.Property(type=GLib.strv_get_type())
    id = GObject.Property(type=str, default='')
    in_production = GObject.Property(type=bool, default=True)
    last_air_date = GObject.Property(type=str, default='')
    last_episode_number = GObject.Property(type=str, default='')
    manual = GObject.Property(type=bool, default=False)
    new_release = GObject.Property(type=bool, default=False)
    next_air_date = release_date = GObject.Property(type=str, default='')
    original_language = GObject.Property(type=LanguageModel)
    original_title = GObject.Property(type=str, default='')
    overview = GObject.Property(type=str, default='')
    poster_path = GObject.Property(type=str, default='')
    recent_change = GObject.Property(type=bool, default=False)
    release_date = GObject.Property(type=str, default='')
    seasons_number = GObject.Property(type=int, default=0)
    seasons = GObject.Property(type=object)
    soon_release = GObject.Property(type=bool, default=False)
    status = GObject.Property(type=str, default='')
    tagline = GObject.Property(type=str, default='')
    title = GObject.Property(type=str, default='')
    watched = GObject.Property(type=bool, default=False)
    notes = GObject.Property(type=str, default='')

    def __init__(self, d=None, t=None):
        super().__init__()

        if d is not None:
            self.add_date = datetime.now()
            self.backdrop_path = self._download_background(d['backdrop_path'])
            self.created_by = self._parse_creators(api_dict=d['created_by'])
            self.episodes_number = d['number_of_episodes']
            self.genres = self._parse_genres(api_dict=d['genres'])
            self.id = d['id']
            self.in_production = d['in_production']
            self.last_air_date = d['last_air_date']
            self.last_episode_number = f"{d['last_episode_to_air']['season_number']}.{d['last_episode_to_air']['episode_number']}"
            self.manual = False
            self.new_release = False
            next_episode_to_air = d['next_episode_to_air']
            if next_episode_to_air == None:
                self.next_air_date = ""
            else:
                self.next_air_date = next_episode_to_air['air_date']
            self.original_language = local.LocalProvider.get_language_by_code(
                d['original_language'])  # type: ignore
            self.original_title = d['original_name']
            self.overview = re.sub(r'\s{2}', ' ', d['overview'])
            # Here the color is also set since it was easy to hook it into the poster download
            self.poster_path, self.color = self._download_poster(
                d['poster_path'])
            self.recent_change = False
            self.release_date = d['first_air_date']
            self.seasons_number = d['number_of_seasons']
            self.seasons = self._parse_seasons(d['seasons'])
            self.status = d['status']
            self.tagline = d['tagline']
            self.title = d['name']
            self.watched = False
            self.activate_notification = self.in_production
            if self.next_air_date != '':
                self.soon_release = datetime.strptime(
                    self.next_air_date, '%Y-%m-%d') < datetime.now() + timedelta(days=6)
            else:
                self.soon_release = False
            self.notes = ''
        else:
            self.activate_notification = t["activate_notification"] # type: ignore
            self.add_date = t["add_date"]  # type: ignore
            self.backdrop_path = t["backdrop_path"]  # type: ignore
            self.created_by = self._parse_creators(
                db_str=t["created_by"])  # type: ignore
            self.color = t["color"]  # type: ignore
            self.episodes_number = t["episodes_number"]  # type: ignore
            self.genres = self._parse_genres(
                db_str=t["genres"])  # type: ignore
            self.id = t["id"]  # type: ignore
            self.in_production = t["in_production"]  # type: ignore
            self.last_air_date = t["last_air_date"]  # type: ignore
            self.last_episode_number = t["last_episode_number"]  # type: ignore
            self.manual = t["manual"]  # type: ignore
            self.new_release = t["new_release"]
            self.next_air_date = t["next_air_date"]
            self.original_language = local.LocalProvider.get_language_by_code(
                t["original_language"])  # type: ignore
            self.original_title = t["original_title"]  # type: ignore
            self.overview = t["overview"]  # type: ignore
            self.poster_path = t["poster_path"]  # type: ignore
            self.recent_change = t["recent_change"]  # type: ignore
            self.release_date = t["release_date"]  # type: ignore
            self.seasons_number = t["seasons_number"]  # type: ignore
            self.soon_release = t["soon_release"]  # type: ignore
            self.status = t["status"]  # type: ignore
            self.tagline = t["tagline"]  # type: ignore
            self.title = t["title"]  # type: ignore
            self.watched = t["watched"]  # type: ignore

            if len(t) == 28:    # type: ignore # TODO: increase this number every time a new field is added
                self.seasons = t["seasons"]  # type: ignore
            else:
                self.seasons = local.LocalProvider.get_all_seasons(
                    self.id)  # type: ignore
                
            self.notes = t["notes"]  # type: ignore

    def _parse_genres(self, api_dict: dict = {}, db_str: str = '') -> List[str]:
        """
        Function to parse genres into a list of strings. Genres are provided by the TMDB API as a dict and are stored in the local db as a comma-separated string.
        Providing both arguments is an error.

        Args:
            from_api (dict): dict from TMDB API
            from_db (str): string from local db

        Returns:
            list of strings
        """

        genres = []

        if api_dict:
            for genre in api_dict:
                genres.append(genre['name'])
            return genres

        if db_str:
            return db_str.split(',')

        return genres

    def _parse_creators(self, api_dict: dict = {}, db_str: str = '') -> List[str]:
        """
        Function to parse the creators into a list of strings. Creators are provided by the TMDB API as a dict and are stored in the local db as a comma-separated string.
        Providing both arguments is an error.

        Args:
            from_api (dict): dict from TMDB API
            from_db (str): string from local db

        Returns:
            list of strings
        """

        creators = []
        if api_dict:
            for creator in api_dict:
                creators.append(creator['name'])
            return creators

        if db_str:
            return db_str.split(',')

        return creators

    def _parse_seasons(self, api_dict: dict) -> List[SeasonModel]:
        """
        Function to parse the seasons data into a list of SeasonModels.

        Args:
            api_dict (dict): dict from TMDB API

        Returns:
            list of SeasonModel
        """

        seasons = []

        for season in api_dict:
            seasons.append(SeasonModel(show_id=self.id, d=season))
        return seasons

    def _download_background(self, path: str) -> str:
        """
        Returns the uri of the background image on the local filesystem, downloading if necessary.

        Args:
            path (str): path to dowload from

        Returns:
            str with the uri of the background image
        """

        if not path:
            return ''

        files = glob.glob(f'{path[1:-4]}.jpg', root_dir=shared.background_dir)
        if files:
            return f'file://{shared.background_dir}/{files[0]}'

        url = f'https://image.tmdb.org/t/p/w500{path}'
        try:
            r = requests.get(url)
            if r.status_code == 200:
                with open(f'{shared.background_dir}{path}', 'wb') as f:
                    f.write(r.content)

                with Image.open(f'{shared.background_dir}{path}') as image:
                    image = (
                        image.convert('RGB')
                        .filter(ImageFilter.GaussianBlur(20))
                    )

                    image.save(f'{shared.background_dir}{path}', 'JPEG')

                return f'file://{shared.background_dir}{path}'
            else:
                return ''
        except (requests.exceptions.ConnectionError, requests.exceptions.SSLError):
            return ''

    def _download_poster(self, path: str) -> Tuple[str, bool]:
        """
        Returns the uri of the poster image on the local filesystem, downloading if necessary.

        Args:
            path (str): path to dowload from

        Returns:
            str with the uri of the poster image
        """

        if not path:
            return f'resource://{shared.PREFIX}/blank_poster.jpg', False

        files = glob.glob(f'{path[1:-4]}.jpg', root_dir=shared.poster_dir)
        if files:
            color = self._compute_badge_color(Path(f'{files[0]}'))
            return f'file://{shared.poster_dir}/{files[0]}', color

        url = f'https://image.tmdb.org/t/p/w500{path}'
        try:
            r = requests.get(url)
            if r.status_code == 200:
                with open(f'{shared.poster_dir}{path}', 'wb') as f:
                    f.write(r.content)
                color = self._compute_badge_color(Path(f'{path}'))
                return f'file://{shared.poster_dir}{path}', color
            else:
                return f'resource://{shared.PREFIX}/blank_poster.jpg', color
        except (requests.exceptions.ConnectionError, requests.exceptions.SSLError):
            return f'resource://{shared.PREFIX}/blank_poster.jpg', False

    def _compute_badge_color(self, path: str) -> bool:
        color_light = False
        im = Image.open(Path(f'{shared.poster_dir}/{path}'))
        box = (im.size[0]-175, 0, im.size[0], 175)
        region = im.crop(box)
        median = ImageStat.Stat(region).median
        if sum(median) < 3 * 128:
            color_light = True

        return color_light



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from gettext import gettext as _

from gi.repository import Adw, GObject, Gtk

import src.providers.local_provider as local

from .. import shared  # type: ignore
from ..models.movie_model import MovieModel
from ..models.series_model import SeriesModel
from ..pages.details_page import DetailsView
from ..widgets.poster_button import PosterButton


@Gtk.Template(resource_path=shared.PREFIX + '/ui/views/content_view.ui')
class ContentView(Adw.Bin):
    """
    This class represents the content grid view.

    Properties:
        None

    Methods:
        refresh(): Causes the view to update its contents

    Signals:
        None
    """

    __gtype_name__ = 'ContentView'

    movie_view = GObject.Property(type=bool, default=True)
    icon_name = GObject.Property(type=str, default='movies')

    _stack = Gtk.Template.Child()
    _updating_status_lbl = Gtk.Template.Child()
    _title_lbl = Gtk.Template.Child()
    _flow_box = Gtk.Template.Child()
    _full_box = Gtk.Template.Child()
    _separated_box = Gtk.Template.Child()
    _unwatched_box = Gtk.Template.Child()
    _unwatched_flow_box = Gtk.Template.Child()
    _watched_box = Gtk.Template.Child()
    _watched_flow_box = Gtk.Template.Child()

    def __init__(self, movie_view: bool):
        super().__init__()
        self.movie_view = movie_view
        self.icon_name = 'movies' if self.movie_view else 'series'

        self._stack.set_visible_child_name('loading')

        self._load_content(self.movie_view)

        self._set_sorting_function()
        self._set_filter_function()

        shared.schema.connect('changed::view-sorting', self._on_sort_changed)
        shared.schema.connect('changed::separate-watched',
                              self._on_separate_watched_changed)
        shared.schema.connect('changed::hide-watched',
                              self._on_hide_watched_changed)
        shared.schema.connect('changed::search-query',
                              self._on_search_enabled_changed)

    def _on_sort_changed(self, pspec: GObject.ParamSpec, user_data: object | None) -> None:
        """
        Callback for the "changed" signal.
        Calls the function to set the sorting function on the FlowBox.

        Args:
            pspec (GObject.ParamSpec): pspec of the changed property
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        self._set_sorting_function()

    def _on_separate_watched_changed(self, pspec: GObject.ParamSpec, user_data: object | None) -> None:
        """
        Callback for the "changed" signal.


        Args:
            pspec (GObject.ParamSpec): pspec of the changed property
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        self.refresh_view()

    def _on_hide_watched_changed(self, pspec: GObject.ParamSpec, user_data: object | None) -> None:
        """
        Callback for the "changed" signal.


        Args:
            pspec (GObject.ParamSpec): pspec of the changed property
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        self._set_filter_function()

    def _load_content(self, movie_view: bool) -> None:
        """
        For each title currently in the db, creates a PosterButton and adds it to the FlowBox.

        Args:
            movie_view (bool): if true it will load movies, otherwise it will load series

        Returns:
            None
        """

        # self._stack.set_visible_child_name('loading')
        if movie_view:
            content = local.LocalProvider.get_all_movies()
        else:
            content = local.LocalProvider.get_all_series()

        if not content:
            self._stack.set_visible_child_name('empty')
            return
        else:
            self._stack.set_visible_child_name('filled')

        for item in content:
            logging.debug(
                f'Created poster button for [{"movie" if self.movie_view else "TV series"}] {item.title}')
            btn = PosterButton(content=item)
            btn.connect('clicked', self._on_clicked)
            if not shared.schema.get_boolean('search-enabled') and shared.schema.get_boolean('separate-watched'):
                if item.watched:
                    self._watched_flow_box.insert(btn, -1)
                else:
                    self._unwatched_flow_box.insert(btn, -1)
            else:
                self._flow_box.insert(btn, -1)

        idx = 0
        while self._flow_box.get_child_at_index(idx):
            self._flow_box.get_child_at_index(idx).set_focusable(False)
            idx += 1

        idx = 0
        while self._watched_flow_box.get_child_at_index(idx):
            self._watched_flow_box.get_child_at_index(idx).set_focusable(False)
            idx += 1

        idx = 0
        while self._watched_flow_box.get_child_at_index(idx):
            self._watched_flow_box.get_child_at_index(idx).set_focusable(False)
            idx += 1

        self._full_box.set_visible(False)
        self._separated_box.set_visible(False)
        self._unwatched_box.set_visible(False)
        self._watched_box.set_visible(False)

        if not shared.schema.get_boolean('search-enabled') and shared.schema.get_boolean('separate-watched'):
            self._separated_box.set_visible(True)
            if self._watched_flow_box.get_child_at_index(0) is not None:
                self._watched_box.set_visible(True)
            if self._unwatched_flow_box.get_child_at_index(0) is not None:
                self._unwatched_box.set_visible(True)
        else:
            self._full_box.set_visible(True)

        # self._stack.set_visible_child_name('filled')

    def refresh_view(self) -> None:
        """
        Causes the view to refresh its contents by replacing the content of the FlowBox.

        Args:
            None

        Returns:
            None
        """

        # self._stack.set_visible_child_name('loading')

        self._flow_box.remove_all()
        self._watched_flow_box.remove_all()
        self._unwatched_flow_box.remove_all()

        self._load_content(self.movie_view)

    def _on_clicked(self, source: Gtk.Widget, content: MovieModel | SeriesModel) -> None:
        """
        Callback for the "clicked" signal.
        Opens the details view for the selected content.

        Args:
            source (Gtk.Widget): widget that emited the signal
            movie (MovieModel): associated movie

        Returns:
            None
        """

        logging.info(
            f'Clicked on [{"movie" if self.movie_view else "TV series"}] {content.title}')
        page = DetailsView(content, self)
        page.connect('deleted', lambda *args: self.refresh_view())
        self.get_ancestor(Adw.NavigationView).push(page)

    def _set_sorting_function(self) -> None:
        """
        Based on the current setting, sets the sorting function of the FlowBox.

        Args:
            None

        Returns;
            None
        """

        match shared.schema.get_string('view-sorting'):
            case 'az':
                self._flow_box.set_sort_func(lambda child1, child2, user_data: (
                    (child1.get_child().title > child2.get_child().title) -
                    (child1.get_child().title < child2.get_child().title)
                ), None)
            case 'za':
                self._flow_box.set_sort_func(lambda child1, child2, user_data: (
                    (child1.get_child().title < child2.get_child().title) -
                    (child1.get_child().title > child2.get_child().title)
                ), None)
            case 'added-date-new':
                self._flow_box.set_sort_func(lambda child1, child2, user_data: (
                    (child1.get_child().content.add_date < child2.get_child().content.add_date) -
                    (child1.get_child().content.add_date >
                     child2.get_child().content.add_date)
                ), None)
            case 'added-date-old':
                self._flow_box.set_sort_func(lambda child1, child2, user_data: (
                    (child1.get_child().content.add_date > child2.get_child().content.add_date) -
                    (child1.get_child().content.add_date <
                     child2.get_child().content.add_date)
                ), None)
            case 'released-date-new':
                self._flow_box.set_sort_func(lambda child1, child2, user_data: (
                    (child1.get_child().content.release_date < child2.get_child().content.release_date) -
                    (child1.get_child().content.release_date >
                     child2.get_child().content.release_date)
                ), None)
            case 'released-date-old':
                self._flow_box.set_sort_func(lambda child1, child2, user_data: (
                    (child1.get_child().content.release_date > child2.get_child().content.release_date) -
                    (child1.get_child().content.release_date <
                     child2.get_child().content.release_date)
                ), None)
        self._flow_box.invalidate_sort()

    def _set_filter_function(self) -> None:
        """
        Based on the current setting, sets the filter function of the FlowBox.

        Args:
            None

        Returns;
            None
        """

        if shared.schema.get_boolean('search-enabled'):
            if shared.schema.get_string('search-mode') == 'title':
                self._flow_box.set_filter_func(
                    lambda child,
                    user_data: (
                        shared.schema.get_string(
                            'search-query').lower() in child.get_child().content.title.lower()
                    ),
                    None)
            else:
                self._flow_box.set_filter_func(
                    lambda child, user_data: (
                        any(
                            shared.schema.get_string(
                                'search-query').title() in genre
                            for genre in child.get_child().content.genres)
                    ),
                    None)
            self._flow_box.invalidate_filter()
            return

        if shared.schema.get_boolean('hide-watched'):
            self._flow_box.set_filter_func(lambda child, user_data: (
                not child.get_child().content.watched), None)
            self._flow_box.invalidate_filter()
            return

        self._flow_box.set_filter_func(lambda child, user_data: True, None)
        self._flow_box.invalidate_filter()

    def _on_search_enabled_changed(self, pspec: GObject.ParamSpec, user_data: object | None) -> None:
        """
        Callback for the "changed" signal.
        Refreshes the view when the search is enabled or disabled.

        Args:
            pspec (GObject.ParamSpec): pspec of the changed property
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        self._title_lbl.set_label(_("Search results") if shared.schema.get_boolean(
            'search-enabled') else _("Your Watchlist"))

        self.refresh_view()

        self._set_filter_function()



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

# isort: off
# autopep8: off
import gi

from src.background_queue import ActivityType, BackgroundActivity, BackgroundQueue

gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')
# isort: on
# autopep: on

import logging
import os
import platform
import subprocess
import sys
from gettext import gettext as _
from gettext import pgettext as C_
from typing import Callable
from datetime import datetime 

from gi.repository import Adw, Gio, GObject, Gtk, GLib

from . import shared  # type: ignore
from .models.search_result_model import SearchResultModel
from .preferences import PreferencesDialog
from .views.content_view import ContentView
from .pages.details_page import DetailsView
from .views.first_run_view import FirstRunView
from .views.db_update_view import DbUpdateView
from .views.main_view import MainView
from .widgets.background_activity_row import BackgroundActivityRow
from .widgets.background_indicator import BackgroundIndicator
from .widgets.episode_row import EpisodeRow
from .widgets.image_selector import ImageSelector
from .widgets.poster_button import PosterButton
from .widgets.search_result_row import SearchResultRow
from .window import TicketboothWindow
from .providers.local_provider import LocalProvider as local

class TicketboothApplication(Adw.Application):
    """The main application singleton class."""

    # Types used in blueprint files
    _custom_widgets = [
        SearchResultModel,
        PosterButton,
        SearchResultRow,
        DetailsView,
        DbUpdateView,
        FirstRunView,
        MainView,
        ContentView,
        EpisodeRow,
        ImageSelector,
        BackgroundIndicator,
        BackgroundActivityRow,
    ]

    def __init__(self):
        super().__init__(application_id=shared.APP_ID, flags=Gio.ApplicationFlags.DEFAULT_FLAGS)

        logging.info('Ticket Booth started')
        logging.info(f'Debug: {shared.DEBUG}')
        logging.debug('Python version: %s', sys.version)
        if os.getenv('FLATPAK_ID') == shared.APP_ID:
            process = subprocess.run(
                ('flatpak-spawn', '--host', 'flatpak', '--version'),
                capture_output=True,
                encoding='utf-8',
                check=False,
            )
            logging.debug('Flatpak version: %s', process.stdout.rstrip())
        logging.debug('Platform: %s', platform.platform())
        if os.name == 'posix':
            for key, value in platform.uname()._asdict().items():
                logging.debug('\t%s: %s', key.title(), value)
        logging.debug('─' * 37)

        self.create_action('quit', lambda *_: self.quit(), ['<primary>q'])
        self.create_action('about', self.on_about_action)
        self.create_action('preferences', self.on_preferences_action, ['<primary>comma'])
        self.create_action('export', self.do_export)
        self.create_action('import', self.do_import)

        for i in self._custom_widgets:
            GObject.type_ensure(i)

    def do_activate(self):
        """
        Called when the application is activated.

        We raise the application's main window, creating it if
        necessary.
        """
        win = self.props.active_window
        if not win:
            win = TicketboothWindow(application=self)

        logging.debug('Window open')

        win.present()

    def on_about_action(self, widget: Gtk.Widget, user_data: object | None):
        """Callback for the app.about action."""

        builder = Gtk.Builder.new_from_resource(shared.PREFIX + '/ui/about_dialog.ui')
        about_dialog = builder.get_object('about_dialog')
        about_dialog.set_application_name(shared.APP_NAME)
        about_dialog.set_application_icon(shared.APP_ID)
        about_dialog.set_version(shared.VERSION)
        about_dialog.add_credit_section('Contributors', [
            # your name <your email>
            # your name website
        ])
        about_dialog.add_legal_section('Movie and TV Series Metadata', 'This product uses the TMDB API but is not endorsed or certified by TMDB.', Gtk.License.CUSTOM, 'All rights belong to their respective owners.')
        logging.debug('About window open')
        about_dialog.present(self.props.active_window)

    def on_preferences_action(self, widget: Gtk.Widget, user_data: object | None):
        """Callback for the app.preferences action."""

        pref_dialog = PreferencesDialog()
        logging.debug('Preferences dialog open')
        pref_dialog.present(self.props.active_window)

    def create_action(self, name: Gtk.Widget, callback: Callable, shortcuts=None):
        """
        Add an application action.

        Args:
            name: the name of the action
            callback: the function to be called when the action is activated
            shortcuts: an optional list of accelerators
        """
        action = Gio.SimpleAction.new(name, None)
        action.connect('activate', callback)
        self.add_action(action)
        if shortcuts:
            self.set_accels_for_action(f'app.{name}', shortcuts)
        logging.debug(f'Created action app.{name} ({shortcuts})')

    def do_export(self, widget: Gtk.Widget, user_data: object | None):
        """
        Callback for the app.export action

        Args:
            widget (Gtk.Widget): the widget that triggered the action
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        logging.info('Export requested')
        logging.info('Opening file dialog to select export location')
        self.dialog = Gtk.FileDialog.new()
        self.dialog.set_modal(True)
        self.dialog.set_initial_name(f"ticketbooth-export-{datetime.now().strftime('%Y-%m-%d-%H:%M:%S')}.zip")
        
        file_filter_store = Gio.ListStore.new(Gtk.FileFilter)
        file_filter = Gtk.FileFilter()
        file_filter.add_suffix("zip")
        file_filter_store.append(file_filter)

        self.dialog.set_filters(file_filter_store)
        self.dialog.save(self.props.active_window, None, self._on_file_save_complete, None)

    def _on_file_save_complete(self,
                               source: Gtk.Widget,
                               result: Gio.AsyncResult,
                               user_data: object | None) -> None:
        """
        Callback for the file dialog.
        Finishes the file selection and, if successfull, asks for confirmation

        Args:
            source (Gtk.Widget): caller widget
            result (Gio.AsyncResult): a Gio.AsyncResult
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """
        try:
            self.archive = self.dialog.save_finish(result)
            
            dialog = Adw.AlertDialog.new(
                heading=_("Confirm Export?"),
                body=_("Exporting your library will create an archive containing all your data. You can import this archive later to restore your library."),
            )
            dialog.add_response('cancel', C_('alert dialog action', '_Cancel'))
            dialog.add_response('export', C_('alert dialog action', '_Export'))
            dialog.set_default_response('export')
            dialog.set_close_response('cancel')
            dialog.set_response_appearance('export', Adw.ResponseAppearance.SUGGESTED)
            dialog.choose(self.props.active_window, None, self._on_export_alert_dialog_choose, None)

        except GLib.GError:
            logging.info("Export cancelled")

    def _on_export_alert_dialog_choose(self,
                                       source: GObject.Object | None,
                                       result: Gio.AsyncResult,
                                       user_data: object | None) -> None:
        """
        Callback for the export confirmation dialog.
        If the user confirms the export, the export process is started.

        Args:
            dialog (Adw.AlertDialog): the dialog that triggered the response
            response (int): the response code

        Returns:
            None
        """
        result = Adw.AlertDialog.choose_finish(source, result)

        if result == 'cancel':
            logging.info("Export cancelled")
            return
        
        BackgroundQueue.add(
            activity=BackgroundActivity(
                activity_type=ActivityType.ADD,
                title=C_('Background activity title', 'Exporting library'),
                task_function=self._export_content_from_db),
            on_done=self._on_export_done
        )
        
    def _export_content_from_db(self, activity: BackgroundActivity) -> None:
        """
        Exports the library to the selected archive.

        Args:
            activity (BackgroundActivity): the calling activity

        Returns:
            None
        """
        if not local.export_data(self.archive.get_path()):
            activity.error()
            logging.info("Export failed")
            
    def _on_export_done(self,
                     source: GObject.Object,
                     result: Gio.AsyncResult,
                     cancellable: Gio.Cancellable,
                     activity: BackgroundActivity):
        """Callback to complete async activity"""

        activity.activity_finish(result, self)
        activity.end()
        logging.info("Export done")


    def do_import(self, new_state: None, source: Gtk.Widget) -> None:
        """
        Callback for the app.import action

        Args:
            new_state (None): the new state of the action
            source (Gtk.Widget): the widget that triggered the action

        Returns:
            None
        """

        logging.info('Import requested')
        logging.info('Opening file dialog to select import location')
        self.dialog = Gtk.FileDialog.new()
        self.dialog.set_modal(True)

        file_filter_store = Gio.ListStore.new(Gtk.FileFilter)
        file_filter = Gtk.FileFilter()
        file_filter.add_suffix("zip")
        file_filter_store.append(file_filter)

        self.dialog.set_filters(file_filter_store)
        self.dialog.open(self.props.active_window, None, self._on_file_open_complete, None)

    def _on_file_open_complete(self,
                               source: Gtk.Widget,
                               result: Gio.AsyncResult,
                               user_data: object | None) -> None:
        """
        Callback for the file dialog.
        Finishes the file selection and, if successfull, asks for confirmation

        Args:
            source (Gtk.Widget): caller widget
            result (Gio.AsyncResult): a Gio.AsyncResult
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        try:
            self.archive = self.dialog.open_finish(result)

            dialog = Adw.AlertDialog.new(
                heading=_("Confirm Import?"),
                body=_("Importing a library will merge it with the current library. This action cannot be undone and in case of conflicts, the new data will be kept."),
            )
            dialog.add_response('cancel', C_('alert dialog action', '_Cancel'))
            dialog.add_response('import', C_('alert dialog action', '_Import'))
            dialog.set_default_response('import')
            dialog.set_close_response('cancel')
            dialog.set_response_appearance('import', Adw.ResponseAppearance.SUGGESTED)
            dialog.choose(self.props.active_window, None, self._on_import_alert_dialog_choose, None)

        except GLib.GError:
            logging.info("Import cancelled")

    def _on_import_alert_dialog_choose(self,
                                       source: GObject.Object | None,
                                       result: Gio.AsyncResult,
                                       user_data: object | None) -> None:
        """
        Callback for the import confirmation dialog.
        If the user confirms the import, the import process is started.

        Args:
            dialog (Adw.AlertDialog): the dialog that triggered the response
            response (int): the response code

        Returns:
            None
        """
        result = Adw.AlertDialog.choose_finish(source, result)

        if result == 'cancel':
            logging.info("Import cancelled")
            return

        BackgroundQueue.add(
            activity=BackgroundActivity(
                activity_type=ActivityType.ADD,
                title=C_('Background activity title', 'Importing library'),
                task_function=self._import_content_to_db),
            on_done=self._on_import_done
        )


    def _import_content_to_db(self, activity: BackgroundActivity) -> None:
        """
        Imports the library from the selected archive.

        Args:
            activity (BackgroundActivity): the calling activity

        Returns:
            None
        """
        if not local.import_data(self.archive.get_path()):
            activity.error()
            logging.info("Import failed")


    def _on_import_done(self,
                     source: GObject.Object,
                     result: Gio.AsyncResult,
                     cancellable: Gio.Cancellable,
                     activity: BackgroundActivity):
        """Callback to complete async activity"""

        activity.activity_finish(result, self)
        activity.end()
        self.props.active_window.activate_action('win.refresh')
        logging.info("Import done")

def main():
    """The application's entry point."""
    app = TicketboothApplication()
    return app.run(sys.argv)



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import glob
import logging
from gettext import gettext as _
from gettext import pgettext as C_

import requests
from gi.repository import Adw, Gio, GLib, GObject, Gtk

from .. import shared  # type: ignore
from ..background_queue import (ActivityType, BackgroundActivity,
                                BackgroundQueue)
from ..providers.local_provider import LocalProvider as local


@Gtk.Template(resource_path=shared.PREFIX + '/ui/widgets/search_result_row.ui')
class SearchResultRow(Gtk.ListBoxRow):
    """
    Widget used to show a search result from TMDB in the search window.

    Properties:
        title (str): a title
        year (str): a release year
        description (str): a description
        poster_path (str): an API endpoint for a poster
        tmdb_id (int): a TMDB id
        media_type (str): a media type
        year_visible (bool): whether or not to show the release year

    Methods:
        None

    Signals:
        None
    """

    __gtype_name__ = 'SearchResultRow'

    _poster_picture = Gtk.Template.Child()
    _poster_spinner = Gtk.Template.Child()
    _media_type_lbl = Gtk.Template.Child()
    _add_btn = Gtk.Template.Child()
    _add_spinner = Gtk.Template.Child()

    tmdb_id = GObject.Property(type=int, default=0)
    title = GObject.Property(type=str, default='')
    year = GObject.Property(type=str, default='')
    description = GObject.Property(type=str, default='')
    poster_path = GObject.Property(type=str, default='')
    media_type = GObject.Property(type=str, default='')
    year_visible = GObject.Property(type=bool, default=False)

    def __init__(self):
        super().__init__()

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for the "map" signal.
        Sets the visibility of the release year, the media type label and the poster to show.
        Additionally calls method in another thread to check if the content is already in db.

        Args:
            user_data (object or None): user data passed to the callback

        Returns:
            None
        """

        logging.debug(
            f'Result row for [{"movie" if self.media_type == "movie" else "TV series"}] {self.title}, {self.year} ({self.poster_path})')

        if self.year:
            self.year_visible = True

        if self.media_type == 'movie':
            self._media_type_lbl.set_label(C_('Category', 'Movie'))
        else:
            self._media_type_lbl.set_label(C_('Category', 'TV Series'))

        self._poster_spinner.set_visible(True)
        self._poster_picture.set_file(self._get_poster_file())
        self._check_in_db()

    def _check_in_db(self) -> None:
        """
        Checks if the content is already in db and disables the 'add' button.

        Args:
            thread_data (object or None): data passed to the thread

        Returns:
            None
        """

        if local.get_movie_by_id(self.tmdb_id) or local.get_series_by_id(self.tmdb_id):
            self._add_btn.set_label(_('Already in your watchlist'))
            self._add_btn.set_icon_name('check-plain')
            self._add_btn.set_sensitive(False)

    def _get_poster_file(self) -> None | Gio.File:
        """
        Get the associated poster image. Files can be retrieved from Internet
        or from local storage if already downloaded in the past. In case no image is found, a blank poster will
        be returned.
        The retrieval is done asynchronously.

        Args:
            None

        Returns:
            None or a Gio.File containing an image
        """
        if self.poster_path:
            Gio.Task.new(self, None, self._on_get_poster_done,
                         None).run_in_thread(self._get_poster_thread)
        else:
            self._poster_spinner.set_visible(False)
            return Gio.File.new_for_uri(f'resource://{shared.PREFIX}/blank_poster.jpg')

    def _on_get_poster_done(self, source_widget: GObject.Object | None, result: Gio.AsyncResult,
                            user_data: object | None) -> None:
        """
        Callback for the async poster retrieval.
        Hides the spinner and shows the poster.

        Args:
            source (GObject.Object or None): the object the asynchronous operation was started with.
            result (Gio.AsyncResult): a Gio.AsyncResult
            user_data (object or None): user data passed to the callback

        Returns:
            None
        """

        poster = self._get_poster_file_finish(result, self)
        self._poster_spinner.set_visible(False)
        self._poster_picture.set_file(poster)

    @Gtk.Template.Callback('_on_add_btn_clicked')
    def _on_add_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Adds a background activity to add content to the local db.

        Args:
            user_data (object or None): user data passed to the callback

        Returns:
            None
        """

        logging.info(
            f'Clicked result for [{"movie" if self.media_type == "movie" else "TV series"}] {self.title}, {self.year}')

        self._add_spinner.set_visible(True)
        self._add_btn.set_sensitive(False)
        BackgroundQueue.add(
            activity=BackgroundActivity(
                activity_type=ActivityType.ADD,
                title=C_('Background activity title',
                         'Add {title}').format(title=self.title),
                task_function=self._add_content_to_db),
            on_done=self._on_add_done)

    def _add_content_to_db(self, activity: BackgroundActivity) -> None:
        """
        Adds the associated title to the corresponding table in the db. Disables the 'add' button when done.

        Args:
            activity (BackgroundActivity): the calling activity

        Returns:
            None
        """
        try:
            local.add_content(id=self.tmdb_id, media_type=self.media_type)
        except ConnectionError:
            activity.set_error(True)

    def _on_add_done(self,
                     source: GObject.Object,
                     result: Gio.AsyncResult,
                     cancellable: Gio.Cancellable,
                     activity: BackgroundActivity):
        """Callback to complete async activity"""

        activity.activity_finish(result, self)        
        self._add_btn.set_label(_('Already in your watchlist'))
        self._add_btn.set_icon_name('check-plain')
        self._add_spinner.set_visible(False)
        
        # Try getting root window and refreshing. If root is gone, use application instance
        if self.get_root():
            self.get_root().activate_action('win.refresh', None)
        else:
            app = Gio.Application.get_default()
            if app:
                app.props.active_window.activate_action('win.refresh')
        
        activity.end()

    def _get_poster_thread(self, task: Gio.Task, source_object: GObject.Object, task_data: object | None,
                           cancelable: Gio.Cancellable | None) -> None:
        """
        Wraper around a blocking function to run it as a non-blocking.

        Args:
            task (Gio.Task): the Gio.Task
            source_object (GObject.Object): task's source object
            task_data (object or None): task's task data
            cancellable (Gio.Cancellable or None): task's Gio.Cancellable, or None

        Returns:
            None
        """

        if task.return_error_if_cancelled():
            return
        outcome = self._get_poster()
        task.return_value(outcome)

    def _get_poster(self) -> Gio.File:
        """
        Returns the poster file from cached data, downloading it if necessary.

        Args:
            None

        Returns:
            Gio.File containing the poster
        """

        files = glob.glob(
            f'{self.poster_path[1:-4]}.jpg', root_dir=shared.cache_dir)
        if files:
            logging.debug(
                f'{self.poster_path}, cache hit: {shared.cache_dir}/{files[0]}')
            return Gio.File.new_for_path(f'{shared.cache_dir}/{files[0]}')
        else:
            url = f'https://image.tmdb.org/t/p/w500{self.poster_path}'
            r = requests.get(url)
            if r.status_code == 200:
                with open(f'{shared.cache_dir}{self.poster_path}', 'wb') as f:
                    f.write(r.content)
            logging.debug(
                f'{self.poster_path}, downloaded to {shared.cache_dir}{self.poster_path}')
            return Gio.File.new_for_path(f'{shared.cache_dir}{self.poster_path}')

    def _get_poster_file_finish(self, result: Gio.AsyncResult, caller: GObject.Object) -> int | Gio.File:
        """
        Finishes the async operation returning the value.

        Args:
            result: a Gio.AsyncResult
            caller: the caller of the task

        Returns:
            int or Gio.File
        """

        if not Gio.Task.is_valid(result, caller):
            return -1
        return result.propagate_value().value



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gio, GObject, Gtk
from pathlib import Path
from .. import shared  # type: ignore
from ..models.movie_model import MovieModel
from ..models.series_model import SeriesModel


@Gtk.Template(resource_path=shared.PREFIX + '/ui/widgets/poster_button.ui')
class PosterButton(Gtk.Box):
    """
    Widget shown in the main view with poster, title, and release year.

    Properties:
        title (str): content's title
        year (str): content's release year
        tmdb_id (str): content's tmdb id
        poster_path (str): content's poster uri

    Methods:
        None

    Signals:
        clicked(content: MovieModel or SeriesModel): emited when the user clicks on the widget
    """

    __gtype_name__ = 'PosterButton'

    _poster_box = Gtk.Template.Child()
    _picture = Gtk.Template.Child()
    _spinner = Gtk.Template.Child()
    _year_lbl = Gtk.Template.Child()
    _status_lbl = Gtk.Template.Child()
    _watched_lbl = Gtk.Template.Child()
    _new_release_badge = Gtk.Template.Child()
    _soon_release_badge = Gtk.Template.Child()
    _watched_badge = Gtk.Template.Child()


    # Properties
    title = GObject.Property(type=str, default='')
    year = GObject.Property(type=str, default='')
    status = GObject.Property(type=str, default='')
    tmdb_id = GObject.Property(type=str, default='')
    poster_path = GObject.Property(type=str, default='')
    watched = GObject.Property(type=bool, default=False)
    content = GObject.Property(type=object, default=None)

    __gsignals__ = {
        'clicked': (GObject.SIGNAL_RUN_FIRST, None, (object,)),
    }

    def __init__(self, content: MovieModel | SeriesModel):
        super().__init__()
        self.activate_notification = content.activate_notification
        self.title = content.title
        self.badge_color_light = content.color
        self.year = content.release_date[0:4] if content.release_date else None
        self.tmdb_id = content.id
        self.poster_path = content.poster_path
        self.watched = content.watched
        self.status = content.status
        self.new_release = content.new_release
        self.soon_release = content.soon_release
        self.recent_change = content.recent_change
        self.content = content

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for the 'map' signal.
        Sets images and hides release year label if not present.

        Args:
            user_data (object or None): data passed to the callback

        Returns:
            None
        """


        self._picture.set_file(Gio.File.new_for_uri(self.poster_path))
        self._spinner.set_visible(False)

        badge_visible = False
        if self.activate_notification:
            if self.recent_change:
                self._poster_box.add_css_class("pulse")
                self._picture.add_css_class("shadow")
            
            if self.new_release:
                self._new_release_badge.set_visible(True)
                badge_visible = True
                if self.badge_color_light:
                    self._new_release_badge.add_css_class("light")
                else:
                    self._new_release_badge.add_css_class("dark")
            elif self.soon_release:
                self._soon_release_badge.set_visible(True)
                badge_visible = True
                if self.badge_color_light:
                    self._soon_release_badge.add_css_class("light")
                else:
                    self._soon_release_badge.add_css_class("dark")

            
        if not self.year:
            self._year_lbl.set_visible(False)
        if self.status == '':
            self._status_lbl.set_visible(False)
        if self.watched and not badge_visible:
            self._watched_badge.set_visible(True)
            if self.badge_color_light:
                self._watched_badge.add_css_class("light")
            else:
                self._watched_badge.add_css_class("dark")

    @Gtk.Template.Callback('_on_poster_btn_clicked')
    def _on_poster_btn_clicked(self, user_data: object | None) -> None:
        self.emit('clicked', self.content)



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import glob
import re
from datetime import datetime, timedelta
from typing import List, Tuple
from pathlib import Path

import requests
from gi.repository import GLib, GObject
from PIL import Image, ImageFilter, ImageStat

import src.providers.local_provider as local

from .. import shared  # type: ignore
from ..models.language_model import LanguageModel


class MovieModel(GObject.GObject):
    """
    This class represents a movie object stored in the db.

    Properties:
        activate_notification (bool): Provide notification if the movie is added to the watchlist before its release date.
        add_date (str): date of addition to the db (ISO format)
        backdrop_path (str): path where the background image is stored
        budget (float): movie budget
        color (bool): color of the poster badges, False being dark
        genres (List[str]): list of genres
        id (str): movie id
        manual (bool): if movie is added manually
        new_release (bool): if the movie has had its release
        original_language (LanguageModel): LanguageModel of the original language
        original_title (str): movie title in original language
        overview (str): movie overview, usually the main plot
        poster_path (str): path where the backgroud poster is stored
        recent_change (bool): indicates that new/soon_release was changed during the last check
        release_date (str): release date in YYYY-MM-DD format
        revenue (float): movie revenue
        runtime (int): movie runtime in minutes
        tagline (str): movie tagline
        soon_release (bool): if the movie has its release soon
        status (str): movie status, usually released or planned
        title (str): movie title
        watched (bool): if the movie has been market as watched
        notes (str): additional notes added by the user

    Methods:
        None

    Signals:
        None
    """

    __gtype_name__ = 'MovieModel'

    activate_notification = GObject.Property(type=bool, default=False)
    add_date = GObject.Property(type=str, default='')
    backdrop_path = GObject.Property(type=str, default='')
    budget = GObject.Property(type=float, default=0)
    color = GObject.Property(type=bool, default=False)
    genres = GObject.Property(type=GLib.strv_get_type())
    id = GObject.Property(type=str, default='')
    manual = GObject.Property(type=bool, default=False)
    new_release = GObject.Property(type=bool, default=False)
    original_language = GObject.Property(type=LanguageModel)
    original_title = GObject.Property(type=str, default='')
    overview = GObject.Property(type=str, default='')
    poster_path = GObject.Property(type=str, default='')
    recent_change = GObject.Property(type=bool, default=False)
    release_date = GObject.Property(type=str, default='')
    revenue = GObject.Property(type=float, default=0)
    runtime = GObject.Property(type=int, default=0)
    soon_release = GObject.Property(type=bool, default=False)
    status = GObject.Property(type=str, default='')
    tagline = GObject.Property(type=str, default='')
    title = GObject.Property(type=str, default='')
    watched = GObject.Property(type=bool, default=False)
    notes = GObject.Property(type=str, default='')

    def __init__(self, d=None, t=None):
        super().__init__()

        if d is not None:
            self.add_date = datetime.now().date()
            self.backdrop_path = self._download_background(
                path=d['backdrop_path'])
            self.budget = d['budget']
            self.genres = self._parse_genres(api_dict=d['genres'])
            self.id = d['id']
            self.manual = False
            self.new_release = False
            self.original_language = local.LocalProvider.get_language_by_code(
                d['original_language'])  # type: ignore
            self.original_title = d['original_title']
            self.overview = re.sub(r'\s{2}', ' ', d['overview'])
            self.poster_path, self.color = self._download_poster(
                path=d['poster_path'], color=False)
            self.recent_change = False
            self.release_date = d['release_date'] if d['release_date'] else None
            self.revenue = d['revenue']
            self.runtime = d['runtime']
            self.soon_release = datetime.strptime(self.release_date, '%Y-%m-%d') < datetime.strptime(self.add_date, '%Y-%m-%d') + timedelta(
                days=14) and datetime.strptime(self.release_date, '%Y-%m-%d') > datetime.strptime(self.add_date, '%Y-%m-%d') if self.release_date and self.add_date else False
            self.status = d['status']
            self.tagline = d['tagline']
            self.title = d['title']
            self.watched = False
            self.activate_notification = datetime.strptime(self.release_date, '%Y-%m-%d') > datetime.strptime(
                # if the release date is in the future activate notifications
                self.add_date, '%Y-%m-%d') if self.release_date and self.add_date else False
            self.notes = ''
        else:
            self.activate_notification = t["activate_notification"] # type: ignore
            self.add_date = t["add_date"]  # type: ignore
            self.backdrop_path = t["backdrop_path"]  # type: ignore
            self.budget = t["budget"]  # type: ignore
            self.color = t["color"] # type: ignore
            self.genres = self._parse_genres(
                db_str=t["genres"])  # type: ignore
            self.id = t["id"]  # type: ignore
            self.manual = t["manual"]  # type:ignore
            self.new_release = t["new_release"] #type: ignore
            self.original_language = local.LocalProvider.get_language_by_code(
                t["original_language"])  # type: ignore
            self.original_title = t["original_title"]  # type: ignore
            self.overview = t["overview"]  # type: ignore
            self.poster_path = t["poster_path"]  # type: ignore
            self.recent_change = t["recent_change"] # type: ignore
            self.release_date = t["release_date"]  # type: ignore
            self.revenue = t["revenue"]  # type: ignore
            self.runtime = t["runtime"]  # type: ignore
            self.soon_release = t["soon_release"] # type: ignore
            self.status = t["status"]  # type: ignore
            self.tagline = t["tagline"]  # type: ignore
            self.title = t["title"]  # type: ignore
            self.watched = t["watched"]  # type:ignore
            self.notes = t["notes"]  # type: ignore

    def _parse_genres(self, api_dict: dict = {}, db_str: str = '') -> List[str]:
        """
        Function to parse genres into a list of strings. Genres are provided by the TMDB API as a dict and are stored in the local db as a comma-separated string.
        Providing both arguments is an error.

        Args:
            from_api (dict): dict from TMDB API
            from_db (str): string from local db

        Returns:
            list of strings
        """

        genres = []

        if api_dict:
            for genre in api_dict:
                genres.append(genre['name'])
            return genres

        if db_str:
            return db_str.split(',')

        return genres

    def _download_background(self, path: str) -> str:
        """
        Returns the uri of the background image on the local filesystem, downloading if necessary.

        Args:
            path (str): path to dowload from

        Returns:
            str with the uri of the background image
        """

        if not path:
            return ''

        files = glob.glob(f'{path[1:-4]}.jpg', root_dir=shared.background_dir)
        if files:
            return f'file://{shared.background_dir}/{files[0]}'

        url = f'https://image.tmdb.org/t/p/w500{path}'
        try:
            r = requests.get(url)
            if r.status_code == 200:
                with open(f'{shared.background_dir}{path}', 'wb') as f:
                    f.write(r.content)

                with Image.open(f'{shared.background_dir}{path}') as image:
                    image = (
                        image.convert('RGB')
                        .filter(ImageFilter.GaussianBlur(20))
                    )

                    image.save(f'{shared.background_dir}{path}', 'JPEG')

                return f'file://{shared.background_dir}{path}'
            else:
                return ''
        except (requests.exceptions.ConnectionError, requests.exceptions.SSLError):
            return ''

    def _download_poster(self, path: str, color: bool) -> Tuple[str, bool]:
        """
        Returns the uri of the poster image on the local filesystem, downloading if necessary.

        Args:
            path (str): path to dowload from

        Returns:
            str with the uri of the poster image
        """

        if not path:
            return (f'resource://{shared.PREFIX}/blank_poster.jpg', False)

        files = glob.glob(f'{path[1:-4]}.jpg', root_dir=shared.poster_dir)
        if files:
            color = self._compute_badge_color(Path(f'{files[0]}'))
            return (f'file://{shared.poster_dir}/{files[0]}', color)

        url = f'https://image.tmdb.org/t/p/w500{path}'
        try:
            r = requests.get(url)
            if r.status_code == 200:
                with open(f'{shared.poster_dir}{path}', 'wb') as f:
                    f.write(r.content)
                color = self._compute_badge_color(Path(f'{path}'))
                return f'file://{shared.poster_dir}{path}', color
            else:
                return f'resource://{shared.PREFIX}/blank_poster.jpg', False
        except (requests.exceptions.ConnectionError, requests.exceptions.SSLError):
            return f'resource://{shared.PREFIX}/blank_poster.jpg', False

    def _compute_badge_color(self, path: str) -> bool:
        color_light = False
        im = Image.open(Path(f'{shared.poster_dir}/{path}'))
        box = (im.size[0]-175, 0, im.size[0], 175)
        region = im.crop(box)
        median = ImageStat.Stat(region).median
        if sum(median) < 3 * 128:
            color_light = True

        return color_light



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from gettext import pgettext as C_

from gi.repository import Adw, Gio, GObject, Gtk

import src.dialogs.edit_season_dialog as dialog

from .. import shared  # type: ignore
from ..models.episode_model import EpisodeModel
from ..pages.edit_episode_page import EditEpisodeNavigationPage
from ..providers.local_provider import LocalProvider as local


@Gtk.Template(resource_path=shared.PREFIX + '/ui/widgets/episode_row.ui')
class EpisodeRow(Adw.PreferencesRow):
    """
    Widget that represents an episode inside the season expander row.

    Properties:
        title (str): episode title
        episode_number (int): episode number in a season
        runtime (int): episode runtime in minutes
        overview (str): episode overview
        still_uri (str): episode still image uri
        editable (bool): whether or not to show delete and edit buttons instead of the "watched" checkbutton
        show_controls (bool): whether or not to show the suffix widgets

    Methods:
        None

    Signals:
        toggled(active: bool): emited when the user toggles the checkbox
    """

    __gtype_name__ = 'EpisodeRow'

    id = GObject.Property(type=str, default='')
    title = GObject.Property(type=str, default='')
    episode_number = GObject.Property(type=int, default=0)
    runtime = GObject.Property(type=int, default=0)
    overview = GObject.Property(type=str, default='')
    still_uri = GObject.Property(type=str, default='')
    editable = GObject.Property(type=bool, default=False)
    show_controls = GObject.Property(type=bool, default=True)
    watched = GObject.Property(type=bool, default=False)
    small_controls = GObject.Property(type=bool, default=False)

    __gsignals__ = {
        'watched-clicked': (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    _still_picture = Gtk.Template.Child()
    _runtime_lbl = Gtk.Template.Child()
    _title_lbl = Gtk.Template.Child()
    _watched_btn = Gtk.Template.Child()

    def __init__(self,
                 episode: EpisodeModel | None = None,
                 id: str = '',
                 title: str = '',
                 episode_number: int = 0,
                 runtime: int = 0,
                 overview: str = '',
                 still_uri: str = '',
                 watched: bool = False,
                 editable: bool = False,
                 show_controls: bool = True,
                 small_controls: bool = False):
        super().__init__()

        if episode:
            self.id = episode.id
            self.title = episode.title
            self.episode_number = episode.number
            self.runtime = episode.runtime
            self.overview = episode.overview
            self.still_uri = episode.still_path
            self.watched = episode.watched
        else:
            self.id = id
            self.title = title
            self.episode_number = episode_number
            self.runtime = runtime
            self.overview = overview
            self.still_uri = still_uri
            self.watched = watched

        self.editable = editable
        self.show_controls = show_controls
        self.small_controls = small_controls

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for "map" signal.
        Sets the still image and title/runtime entries.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        if not self.editable and self.show_controls:
            self.watched = local.get_episode_by_id(
                self.id).watched  # type: ignore

        if self.small_controls:
            self._still_picture.set_visible(False)
        else:
            self._still_picture.set_file(Gio.File.new_for_uri(self.still_uri))
            
        self._title_lbl.set_text(f'{self.episode_number}. {self.title}')
        self._runtime_lbl.set_text(self._format_runtime(self.runtime))

        if self.small_controls:
            if self.watched:
                self._watched_btn.set_icon_name('check-plain')
            else:
                self._watched_btn.set_icon_name('watchlist')
        else:
            btn_content = Adw.ButtonContent()
            if self.watched:
                btn_content.set_label(_('Watched'))
                btn_content.set_icon_name('check-plain')
            else:
                btn_content.set_label(_('Mark as Watched'))
                btn_content.set_icon_name('watchlist')
            self._watched_btn.set_child(btn_content)

    @Gtk.Template.Callback('_on_watched_btn_clicked')
    def _on_watched_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Toggles the watched property of the associated episode in the db. The "watched-clicked" signal is emited after this callback is completed.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        local.mark_watched_episode(self.id, not self.watched)
        self.watched = not self.watched

        if self.small_controls:
            if self.watched:
                self._watched_btn.set_icon_name('check-plain')
            else:
                self._watched_btn.set_icon_name('watchlist')
        else:
            btn_content = Adw.ButtonContent()
            if self.watched:
                btn_content.set_label(_('Watched'))
                btn_content.set_icon_name('check-plain')
            else:
                btn_content.set_label(_('Mark as Watched'))
                btn_content.set_icon_name('watchlist')
            self._watched_btn.set_child(btn_content)

        self.emit('watched-clicked')

    def _format_runtime(self, runtime: int) -> str:
        """
        Formats the runtime in hours and minutes.

        Args:
            runtime (str): runtime in minutes

        Returns:
            formatted duration string in hours and minutes
        """

        h, m = divmod(runtime, 60)

        if h > 0:

            # TRANSLATORS: {h} and {m} are the runtime hours and minutes respectively
            return _('{h}h {m}min').format(h=h, m=m)
        else:

            # TRANSLATORS: {m} is the runtime minutes
            return _('{m}min').format(m=m)

    @Gtk.Template.Callback('_on_edit_btn_clicked')
    def _on_edit_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Shows the "edit episode" page.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        edit_episode_page = EditEpisodeNavigationPage(title=self.title,
                                                      episode_number=self.episode_number,
                                                      runtime=self.runtime,
                                                      overview=self.overview,
                                                      still_uri=self.still_uri,
                                                      watched=self.watched)
        edit_episode_page.connect('edit-saved', self._on_episode_saved)
        self.get_ancestor(Adw.NavigationView).push(edit_episode_page)

    def _on_episode_saved(self,
                          source: Gtk.Widget,
                          title: str,
                          episode_number: int,
                          runtime: int,
                          overview: str,
                          still_uri: str,
                          watched: bool) -> None:
        """
        Callback for "edit-saved" signal.
        Appends the recieved data as a tuple in the episodes list after removing the changed one and updates the ui.

        Args:
            source (Gtk.Widget): caller widget
            title (str): episode title
            episode_number (int): episode number
            runtime (int): episode runtime in minutes
            overview (str): episode overview
            still_uri (str): episode still uri

        Returns:
            None
        """

        parent = self.get_ancestor(dialog.EditSeasonDialog)

        old_episode = parent.get_episode(self.title,
                                         self.episode_number,
                                         self.runtime,
                                         self.overview,
                                         self.still_uri)
        parent._episodes.remove(old_episode)
        parent.update_episodes_ui()

        parent._episodes.append(
            (title, episode_number, runtime, overview, still_uri, watched))

        parent.update_episodes_ui()

    @Gtk.Template.Callback('_on_delete_btn_clicked')
    def _on_delete_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Asks the user for a confirmation after a delete request.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        # TRANSLATORS: {title} is the showed content's title
        dialog = Adw.AlertDialog.new(
            heading=C_('alert dialog heading', 'Delete {title}?').format(
                title=f'{self.episode_number}. {self.title}'),
            body=C_('alert dialog body',
                    'All changes to this episode will be lost.')
        )
        dialog.add_response('cancel', C_('alert dialog action', '_Cancel'))
        dialog.add_response('delete', C_('alert dialog action', '_Delete'))
        dialog.set_default_response('delete')
        dialog.set_close_response('cancel')
        dialog.set_response_appearance(
            'delete', Adw.ResponseAppearance.DESTRUCTIVE)
        dialog.choose(self, None, self._on_alert_dialog_choose, None)

    def _on_alert_dialog_choose(self,
                                source: GObject.Object | None,
                                result: Gio.AsyncResult,
                                user_data: object | None) -> None:
        """
        Callback for the alert dialog.
        Finishes the async operation and retrieves the user response. If the later 
        is positive, deletes the episode from the list and updates the ui.

        Args:
            source (Gtk.Widget): object that started the async operation
            result (Gio.AsyncResult): a Gio.AsyncResult
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        result = Adw.AlertDialog.choose_finish(source, result)
        if result == 'cancel':
            return

        parent = self.get_ancestor(dialog.EditSeasonDialog)
        old_episode = parent.get_episode(self.title,
                                         self.episode_number,
                                         self.runtime,
                                         self.overview,
                                         self.still_uri)
        parent._episodes.remove(old_episode)
        parent.update_episodes_ui()

    def set_watched_btn(self, watched: bool) -> None:
        """
        Updates the watched button label and icon.

        Args:
            watched (bool): status to set

        Returns:
            None
        """

        if self.small_controls:
            if watched:
                self._watched_btn.set_icon_name('check-plain')
            else:
                self._watched_btn.set_icon_name('watchlist')
        else:
            btn_content = Adw.ButtonContent()
            if watched:
                btn_content.set_label(_('Watched'))
                btn_content.set_icon_name('check-plain')
            else:
                btn_content.set_label(_('Mark as Watched'))
                btn_content.set_icon_name('watchlist')
            self._watched_btn.set_child(btn_content)




# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from gettext import pgettext as C_
from typing import List

from gi.repository import Adw, Gio, GObject, Gtk

import src.dialogs.add_manual_dialog as dialog

from .. import shared  # type: ignore
from ..dialogs.edit_season_dialog import EditSeasonDialog
from ..widgets.episode_row import EpisodeRow


@Gtk.Template(resource_path=shared.PREFIX + '/ui/widgets/season_expander.ui')
class SeasonExpander(Adw.ExpanderRow):
    """
    This class represents a season in the manual add window.

    Properties:
        season_title (str): season title
        poster_uri (str): season poster uri

    Methods:
        None

    Signals:
        None
    """

    __gtype_name__ = 'SeasonExpander'

    season_title = GObject.Property(type=str)
    poster_uri = GObject.Property(type=str)
    episodes = GObject.Property(type=object)

    _poster = Gtk.Template.Child()

    def __init__(self,
                 season_title: str = _('Season'),
                 poster_uri: str = f'resource://{shared.PREFIX}/blank_poster.jpg',
                 episodes: List | None = None):

        super().__init__()

        self.season_title = season_title
        self.poster_uri = poster_uri
        self.episodes = episodes if episodes else []

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for "map" signal.
        Sets the poster image and populates the episode data.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        self._poster.set_file(Gio.File.new_for_uri(self.poster_uri))

        for episode in self.episodes:
            self.add_row(EpisodeRow(title=episode[0],
                                    episode_number=episode[1],
                                    runtime=episode[2],
                                    overview=episode[3],
                                    still_uri=episode[4],
                                    watched=episode[5],
                                    show_controls=False)
                         )

    @Gtk.Template.Callback('_on_edit_btn_clicked')
    def _on_edit_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Shows the "edit season" window.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        dialog = EditSeasonDialog(self.get_ancestor(Gtk.Window), title=self.season_title,
                                  poster_uri=self.poster_uri, episodes=self.episodes)
        dialog.connect('edit-saved', self._on_edit_saved)
        dialog.present()

    def _on_edit_saved(self,
                       source: Gtk.Widget,
                       title: str,
                       poster_uri: str,
                       episodes: List[tuple]) -> None:
        """
        Callback for "edit-saved" signal.
        Appends the recieved data as a tuple in the seasons list after removing the changed one and updates the ui.

        Args:
            source (Gtk.Widget): caller widget
            title (str): season title
            poster_uri (str): season poster uri
            episodes (List[tuple]): episodes in season

        Returns:
            None
        """

        parent_dialog = self.get_ancestor(dialog.AddManualDialog)

        old_season = parent_dialog.get_season(self.season_title,
                                              self.poster_uri,
                                              self.episodes)
        parent_dialog.seasons.remove(old_season)

        parent_dialog.seasons.append((title, poster_uri, episodes))

        parent_dialog.update_seasons_ui()

    @Gtk.Template.Callback('_on_delete_btn_clicked')
    def _on_delete_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for the "clicked" signal.
        Asks the user for a confirmation after a delete request.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        # TRANSLATORS: {title} is the showed content's title
        dialog = Adw.AlertDialog.new(
            heading=C_('alert dialog heading', 'Delete {title}?').format(
                title=self.season_title),
            body=C_('alert dialog body',
                    'This season contains unsaved metadata.')
        )
        dialog.add_response('cancel', C_('alert dialog action', '_Cancel'))
        dialog.add_response('delete', C_('alert dialog action', '_Delete'))
        dialog.set_default_response('delete')
        dialog.set_close_response('cancel')
        dialog.set_response_appearance(
            'delete', Adw.ResponseAppearance.DESTRUCTIVE)
        dialog.choose(self, None, self._on_alert_dialog_choose, None)

    def _on_alert_dialog_choose(self, source: GObject.Object | None, result: Gio.AsyncResult, user_data: object | None) -> None:
        """
        Callback for the alert dialog.
        Finishes the async operation and retrieves the user response. If the later is positive, delete the content from the db.

        Args:
            source (Gtk.Widget): object that started the async operation
            result (Gio.AsyncResult): a Gio.AsyncResult
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        result = Adw.AlertDialog.choose_finish(source, result)
        if result == 'cancel':
            return

        parent_dialog = self.get_ancestor(dialog.AddManualDialog)
        old_season = parent_dialog.get_season(
            self.season_title, self.poster_uri, self.episodes)
        parent_dialog.seasons.remove(old_season)
        parent_dialog.update_seasons_ui()



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import locale
import logging
import os
from gettext import gettext as _

from gi.repository import Adw, Gio, GLib, GObject, Gtk

from .. import shared  # type: ignore
from ..models.language_model import LanguageModel
from ..providers.local_provider import LocalProvider as local
from ..providers.tmdb_provider import TMDBProvider as tmdb


@Gtk.Template(resource_path=shared.PREFIX + '/ui/views/first_run_view.ui')
class FirstRunView(Adw.Bin):
    """
    This class represents the initial setup the app needs to offer full fuctionality.

    Properties:
        None

    Methods:
        None

    Signals:
        exit: emited when the view has completed its operations, either the required data was successfully downloaded \
              or the user requested "offline mode"
    """

    __gtype_name__ = 'FirstRunView'

    _heading_lbl = Gtk.Template.Child()
    _status_lbl = Gtk.Template.Child()
    _offline_btn = Gtk.Template.Child()
    _retry_check_btn = Gtk.Template.Child()

    __gsignals__ = {
        'exit': (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    _cancellable = Gio.Cancellable.new()

    def __init__(self):
        super().__init__()

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for "map" signal.
        Creates the directories and tables in the local db, sets the tmdb results language based on the locale, and attempts to download required data if connected to the Internet.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        logging.info('First run setup in progress...')

        for path in [shared.background_dir, shared.poster_dir, shared.series_dir]:
            if not os.path.exists(path):
                os.makedirs(path)
                logging.info(f'[setup] Created folder {path}')

        local.create_tables()
        logging.info('[setup] Created db tables')

        language = locale.getdefaultlocale()[0].lower()[:2]  # type: ignore
        shared.schema.set_string('tmdb-lang', language)
        logging.info(f'[Setup] Set TMDB language to {language}')

        self._update_ui(need_download=True)
        Gio.NetworkMonitor.get_default().can_reach_async(
            Gio.NetworkAddress.parse_uri('https://api.themoviedb.org', 80),
            self._cancellable,
            self._on_first_reach_done,
            None
        )
        logging.info('[Setup] Checking network connection...')

    def _update_ui(self, need_download: bool) -> None:
        """
        Updates the UI strings to reflect its state:
            - if need_download is TRUE, the UI tells the user that the Internet is needed and some data will be
              downloaded. Additionally it shows options to proceed offline.
            - if need_download is FALSE, the UI tells the user a generic message.

        Args:
            need_download (bool): selected messages to be shown as stated above.

        Returns:
            None
        """

        if need_download:
            self._heading_lbl.set_label(_('Waiting for Network…'))
            self._status_lbl.set_label(
                _("For a complete experience, a download of 15 KB is required. However, if you are not connected to the Internet or don't want to wait, you can skip this step and continue offline without some features."))
            self._offline_btn.set_visible(True)
            self._retry_check_btn.set_visible(True)
        else:
            self._heading_lbl.set_label(_('Getting things ready…'))
            self._status_lbl.set_label(_('Downloading data'))
            self._offline_btn.set_visible(False)
            self._retry_check_btn.set_visible(False)

    def _on_first_reach_done(self, source: GObject.Object | None, result: Gio.AsyncResult, data: object | None) -> None:
        """
        Callback for asynchronous network check, reached after the first call.
        If the network is available, proced with the download, otherwise keep checking every second until it becomes
        available or the user goes in offline mode.

        Args:
            source (GObject.Object or None): the object the asynchronous operation was started with.
            result (Gio.AsyncResult): a Gio.AsyncResult
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        try:
            network = Gio.NetworkMonitor.get_default().can_reach_finish(result)
        except GLib.Error:
            network = None

        if network:
            logging.info('[Setup] Network present, staring download')
            self._download_languages()
        else:
            self._has_network = False
            logging.error(
                '[Setup] Network not present, retrying in 10 seconds')
            # Continue checking in a separate thread
            GLib.Thread.new(None, self._loop_check_network)

    def _loop_check_network(self) -> None:
        """
        Function run in a separate thread to continously check for network connection every second until the operation
        is cancelled or the network is restored.
        If the network is restored, the download is completed and the thread killed.

        Args:
            None

        Returns:
            None
        """

        while not (self._cancellable.is_cancelled() or self._has_network):
            GLib.usleep(10000000)
            Gio.NetworkMonitor.get_default().can_reach_async(
                Gio.NetworkAddress.parse_uri('https://api.themoviedb.org', 80),
                self._cancellable,
                self._on_loop_reach_done,
                None
            )

        if self._has_network:
            logging.info('[Setup] Network present, staring download')
            self._download_languages()

        GLib.Thread.exit()

    def _on_loop_reach_done(self, source: GObject.Object | None, result: Gio.AsyncResult, user_data: object | None) -> None:
        """
        Callback for asynchronous network check, reached during the looped check.
        It sets the network presence flag based on the current condition.

        Args:
            source (GObject.Object or None): the object the asynchronous operation was started with.
            result (Gio.AsyncResult): a Gio.AsyncResult
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        try:
            self._has_network = Gio.NetworkMonitor.get_default().can_reach_finish(result)
        except GLib.Error:
            self._has_network = False
            logging.error(
                '[Setup] Network not present, retrying in 10 seconds')

    def _download_languages(self) -> None:
        """
        Completes the downlaod, stores the data in the db and sets the relevant GSettings.

        Args:
            None

        Results:
            None
        """

        self._update_ui(need_download=False)

        languages = tmdb.get_languages()
        for lang in languages:
            local.add_language(LanguageModel(lang))

        shared.schema.set_boolean('first-run', False)
        shared.schema.set_boolean('offline-mode', False)
        shared.schema.set_boolean('onboard-complete', True)
        logging.info('[Setup] First setup complete')
        self.emit('exit')

    @Gtk.Template.Callback('_on_offline_btn_clicked')
    def _on_offline_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Stops the background network check and sets the app in offline mode. An option to retry on next launch is
        provided.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        self._cancellable.cancel()

        shared.schema.set_boolean('offline-mode', True)
        logging.info('[Setup] Offline mode enabled')
        if not self._retry_check_btn.get_active():
            shared.schema.set_boolean('first-run', False)
            logging.info('[Setup] First setup partially complete')

        logging.info('[Setup] Setup not completed, retrying on next run')
        self.emit('exit')



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import re

from gi.repository import GObject


class SearchResultModel(GObject.GObject):
    """
    This class represents the object returned from the TMDB search endpoint.

    Properties:
        title (str): content's title
        year (str): content's release year
        description (str): content's description
        poster_path (str): API endpoint for the content poster
        tmdb_id (int): content's unique id in the API
        media_type (str): content's media type

    Methods:
        None

    Signals:
        None
    """

    __gtype_name__ = 'SearchResultModel'

    title = GObject.Property(type=str, default='')
    year = GObject.Property(type=str, default='')
    description = GObject.Property(type=str, default='')
    poster_path = GObject.Property(type=str, default='')
    tmdb_id = GObject.Property(type=int, default=0)
    media_type = GObject.Property(type=str, default='')

    def __init__(self, d=None):
        super().__init__()

        if d is not None:
            self.tmdb_id = d['id']
            self.poster_path = d['poster_path']
            self.description = re.sub(r'\s{2}', ' ', d['overview'])

            if d['media_type'] == 'movie':
                self.media_type = d['media_type']
                self.title = d['title']
                self.year = d['release_date'][0:4]
            elif d['media_type'] == 'tv':
                self.media_type = d['media_type']
                self.title = d['name']
                self.year = d['first_air_date'][0:4]



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os

import tmdbsimple as tmdb

from .. import shared  # type: ignore


class TMDBProvider:
    """
    This class provides methods to interface with the TMDB API.

    Properties:
        None

    Methods:
        search(query: str, lang: str or None): Searches the API for the given query
        get_languages(): Retrieves all available languages usable with the API
        get_movie(id: int, lang: str): Retrieves general information about a movie.
        get_serie(id: int, lang: str): Retrieves general information about a tv series.
        get_season_episodes(id: int, series:int, lang: str): Retrieves information about the episodes in a season.
    """

    if shared.schema.get_boolean('use-own-tmdb-key'):
        tmdb.API_KEY = shared.schema.get_string('own-tmdb-key')
    else:
        tmdb.API_KEY = os.environ.get('TMDB_KEY')

    def __init__(self):
        super().__init__()

    @staticmethod
    def search(query: str, lang: str | None = None) -> dict:
        """
        Searches the API for the given query.

        Args:
            query (str): a query to lookup
            lang (str or None): the prefered language for the results (ISO 639-1 format)

        Returns:
            dict containg the API result.
        """

        if not lang:
            lang = shared.schema.get_string('tmdb-lang')

        return tmdb.Search().multi(query=query, language=lang, include_adult=False)

    @staticmethod
    def get_languages() -> dict:
        """
        Retrieves all available languages usable with the API

        Args:
            None

        Returns:
            dict containg the API result.
        """

        return tmdb.Configuration().languages()

    @staticmethod
    def get_movie(id: int, lang: str | None = None) -> dict:
        """
        Retrieves general information about the movie with the provided id.

        Args:
            id (int): id of the movie
            lang (str): the prefered language for the results (optional)

        Returns:
            dict containg the API result.
        """
        if not lang:
            lang = shared.schema.get_string('tmdb-lang')

        return tmdb.Movies(id).info(language=lang)

    @staticmethod
    def get_serie(id: int, lang: str | None = None) -> dict:
        """
        Retrieves general information about the tv series with the provided id.

        Args:
            id (int): id of the tv series
            lang (str): the prefered language for the results (optional)

        Returns:
            dict containg the API result
        """
        if not lang:
            lang = shared.schema.get_string('tmdb-lang')
            
        return tmdb.TV(id).info(language=lang)

    @staticmethod
    def get_season_episodes(id: int, season: int, lang: str | None = None) -> dict:
        """
        Retrieves information about the episodes in a season for the specified tv series.

        Args:
            id (int): id of the tv series
            season (int): season number
            lang (str): the prefered language for the results (optional)

        Returns:
            dict containg the API result.
        """

        if not lang:
            lang = shared.schema.get_string('tmdb-lang')

        return tmdb.TV_Seasons(id, season).info(language=lang)['episodes']

    @staticmethod
    def set_key(key: str) -> None:
        """
        Sets the API in use.

        Args:
            key (str): key to use

        Returns:
            None
        """

        tmdb.API_KEY = key

    @staticmethod
    def get_key() -> str:
        """
        Gets the API in use.

        Args:
            None

        Returns:
            str with the key in use
        """

        return tmdb.API_KEY

    @staticmethod
    def get_builtin_key() -> str:
        """
        Gets the builtin API key.

        Args:
            None

        Returns:
            str with the builtin key
        """

        return os.environ.get('TMDB_KEY')  # type: ignore



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import locale
import logging
import os
import sqlite3
import time
from datetime import datetime, timedelta
from gettext import gettext as _

from gi.repository import Adw, Gio, GLib, GObject, Gtk

from .. import shared  # type: ignore
from ..models.movie_model import MovieModel
from ..models.series_model import SeriesModel
from ..providers.local_provider import LocalProvider as local
from ..providers.tmdb_provider import TMDBProvider as tmdb

@Gtk.Template(resource_path=shared.PREFIX + '/ui/views/db_update_view.ui')
class DbUpdateView(Adw.Bin):
    """
    This class represents the initial setup the app needs to offer full fuctionality.

    Properties:
        None

    Methods:
        None

    Signals:
        exit: emited when the view has completed its operations, either the required data was successfully downloaded \
              or the user requested "offline mode"
    """

    __gtype_name__ = 'DbUpdateView'

    _progress_bar = Gtk.Template.Child()
    _update_progress_page = Gtk.Template.Child()
    _notification_question = Gtk.Template.Child()
    _activate_btn = Gtk.Template.Child()
    _deactivate_btn = Gtk.Template.Child()
    _carousel = Gtk.Template.Child()
    _offline_btn = Gtk.Template.Child()
    _retry_check_btn = Gtk.Template.Child()

    __gsignals__ = {
        'exit': (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    _cancellable = Gio.Cancellable.new()

    def __init__(self):
        super().__init__()

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for "map" signal.
        Creates the directories and tables in the local db, sets the tmdb results language based on the locale, and attempts to download required data if connected to the Internet.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """
        

        logging.info('Updating Database...')

        local.update_movies_table()
        local.update_series_table()

        logging.info('Added new Columns to Database')

        if shared.schema.get_boolean('offline-mode'):
            self.emit('exit')
        logging.info('Refetching Data from TMDB...')
        self.task = Gio.Task.new()
        self.task.run_in_thread(
            lambda*_:self._fetch_data_from_tmdb()
        )
       

    @Gtk.Template.Callback('_on_offline_btn_clicked')
    def _on_offline_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Stops the background network check and sets the app in offline mode. An option to retry on next launch is
        provided.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        self._cancellable.cancel()

        shared.schema.set_boolean('offline-mode', True)
        logging.info('[Update] Offline mode enabled')
        if not self._retry_check_btn.get_active():
            shared.schema.set_boolean('db-needs-update', False)
            logging.info('[Update] Database update partially complete')

        logging.info('[Update] Database update not completed, retrying on next run')
        self.emit('exit')

    def _fetch_data_from_tmdb(self) :
        with sqlite3.connect(shared.db) as connection: 

            movies = local.get_all_movies()
            series = local.get_all_series()
            
            total = len(movies) + len(series)
            counter = 0
            for movie in movies:
                if not movie.manual:
                    new_movie = MovieModel(tmdb.get_movie(movie.id))
                    #Check if the soon_release flag should be set, that is the case if the movie is set to release in less than 14 days
                    if len(movie.release_date) > 0 and datetime.strptime(movie.release_date, '%Y-%m-%d') < datetime.now() + timedelta(days=14): # TODO make this a variable and sync with main_view.py
                        #Writing to the local db since update_movie updates entry of the local db
                        local.set_soon_release_status(movie.id, True, True)
                    local.update_movie(movie, new_movie)
                    counter += 1
                    self._progress_bar.set_fraction(counter/total)

            for serie in series:
                if not serie.manual:
                    new_serie = SeriesModel(tmdb.get_serie(serie.id))
                     #Check if the soon_release flag should be set, that is the case if the next episode is less than 7 days away
                    compare_date = new_serie.next_air_date
                    if len(compare_date) > 0 and datetime.strptime(compare_date, '%Y-%m-%d') < datetime.now() + timedelta(days=7): # TODO make this a variable and sync with main_view.py
                        #Writing to variable since update_series deletes local db entry and creates a new entry with flags from argument
                        serie.soon_release = True
                    local.update_series(serie, new_serie)
                    counter += 1
                    self._progress_bar.set_fraction(counter/total)

        logging.info('Fetched Data from TMDB.')

        #Go to next page in the carousel
        index = int(self._carousel.get_position())
        next_page = self._carousel.get_nth_page(index + 1)
        self._carousel.scroll_to(next_page, True)
          
    @Gtk.Template.Callback('_on_deactivate_btn_clicked')
    def _on_btn_deactivate_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Sets all activate_notification to false in local db and exits carousel

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """
        local.reset_activate_notification()
        shared.schema.set_boolean('db-needs-update', False)
        logging.debug(f'db update dialog: confim, deactivate notifications')
        self.emit('exit')

    @Gtk.Template.Callback('_on_activate_btn_clicked')
    def _on_btn_activate_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Exits carousel.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """
        shared.schema.set_boolean('db-needs-update', False)
        self.emit('exit')